############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
#
# written in python3, (c) 2019-2021 by mworion
# Licence APL2.0
#
###########################################################
import base.packageConfig as pConf

# standard libraries
from datetime import datetime
import gc

# external packages
from PyQt5.QtGui import QPixmap
import PyQt5.QtCore
import PyQt5.QtWidgets
from PyQt5.QtTest import QTest

# local import
if pConf.isAvailable:
    from gui.extWindows.simulatorW import SimulatorWindow
    from gui.extWindows.keypadW import KeypadWindow

from gui.utilities.toolsQtWidget import MWidget
from gui.extWindows.messageW import MessageWindow
from gui.extWindows.hemisphereW import HemisphereWindow
from gui.extWindows.measureW import MeasureWindow
from gui.extWindows.imageW import ImageWindow
from gui.extWindows.satelliteW import SatelliteWindow
from gui.extWindows.analyseW import AnalyseWindow
from gui.widgets.main_ui import Ui_MainWindow
from gui.mainWmixin.tabMount import Mount
from gui.mainWmixin.tabEnviron import Environ
from gui.mainWmixin.tabAlmanac import Almanac
from gui.mainWmixin.tabModel import Model
from gui.mainWmixin.tabBuildPoints import BuildPoints
from gui.mainWmixin.tabManageModel import ManageModel
from gui.mainWmixin.tabSatellite import Satellite
from gui.mainWmixin.tabMinorPlanetTime import MinorPlanetTime
from gui.mainWmixin.tabRelay import Relay
from gui.mainWmixin.tabTools import Tools
from gui.mainWmixin.tabPower import Power
from gui.mainWmixin.tabSettDevice import SettDevice
from gui.mainWmixin.tabSettMount import SettMount
from gui.mainWmixin.tabSettHorizon import SettHorizon
from gui.mainWmixin.tabSettImaging import SettImaging
from gui.mainWmixin.tabSettDome import SettDome
from gui.mainWmixin.tabSettParkPos import SettParkPos
from gui.mainWmixin.tabSettRelay import SettRelay
from gui.mainWmixin.tabSettMisc import SettMisc


class MainWindow(
    MWidget,
    SettMisc,
    Mount,
    Environ,
    Almanac,
    Model,
    BuildPoints,
    ManageModel,
    Satellite,
    MinorPlanetTime,
    Relay,
    Power,
    Tools,
    SettDevice,
    SettMount,
    SettHorizon,
    SettImaging,
    SettDome,
    SettParkPos,
    SettRelay,
):
    """
    the main window class handles the main menu as well as the show and no show part of
    any other window. all necessary processing for functions of that gui will be linked
    to this class. therefore window classes will have a threadpool for managing async
    processing if needed.
    """

    __all__ = [
        "MainWindow",
    ]

    def __init__(self, app):
        super().__init__()

        self.app = app
        self.threadPool = app.threadPool
        self.deviceStat = app.deviceStat
        self.uiWindows = app.uiWindows
        self.setAttribute(PyQt5.QtCore.Qt.WA_DeleteOnClose)

        # load and init the gui
        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)
        self.initUI()
        self.setWindowTitle(f"MountWizzard4 - v{self.app.__version__}")

        # link cross widget gui signals as all ui widgets have to be present
        self.uiWindows["showMessageW"] = {
            "button": self.ui.openMessageW,
            "classObj": None,
            "name": "MessageDialog",
            "class": MessageWindow,
        }
        self.uiWindows["showHemisphereW"] = {
            "button": self.ui.openHemisphereW,
            "classObj": None,
            "name": "HemisphereDialog",
            "class": HemisphereWindow,
        }
        self.uiWindows["showImageW"] = {
            "button": self.ui.openImageW,
            "classObj": None,
            "name": "ImageDialog",
            "class": ImageWindow,
        }
        self.uiWindows["showMeasureW"] = {
            "button": self.ui.openMeasureW,
            "classObj": None,
            "name": "MeasureDialog",
            "class": MeasureWindow,
        }
        self.uiWindows["showSatelliteW"] = {
            "button": self.ui.openSatelliteW,
            "classObj": None,
            "name": "SatelliteDialog",
            "class": SatelliteWindow,
        }
        self.uiWindows["showAnalyseW"] = {
            "button": self.ui.openAnalyseW,
            "classObj": None,
            "name": "AnalyseDialog",
            "class": AnalyseWindow,
        }
        if pConf.isSimulator and pConf.isAvailable:
            self.uiWindows["showSimulatorW"] = {
                "button": self.ui.mountConnected,
                "classObj": None,
                "name": "SimulatorDialog",
                "class": SimulatorWindow,
            }
        if pConf.isAvailable:
            self.uiWindows["showKeypadW"] = {
                "button": self.ui.openKeypadW,
                "classObj": None,
                "name": "KeypadDialog",
                "class": KeypadWindow,
            }

        self.deviceStatGui = {
            "dome": self.ui.domeConnected,
            "camera": self.ui.cameraConnected,
            "environOverall": self.ui.environConnected,
            "astrometry": self.ui.astrometryConnected,
            "mount": self.ui.mountConnected,
        }

        self.mwSuper("__init__")

        # polarPlot ui instance has to be defined central, not in the mixins
        self.modelPositionPlot = self.embedMatplot(self.ui.modelPosition)
        self.errorAscendingPlot = self.embedMatplot(self.ui.errorAscending)
        self.errorDistributionPlot = self.embedMatplot(self.ui.errorDistribution)
        self.twilight = self.embedMatplot(self.ui.twilight)
        self.modelPositionPlot.figure.canvas.mpl_connect(
            "button_press_event", self.onMouseEdit
        )

        # connect signals for refreshing the gui
        self.app.mount.signals.pointDone.connect(self.updateStatusGUI)
        self.app.mount.signals.mountUp.connect(self.updateMountConnStat)
        self.app.mount.signals.settingDone.connect(self.updateMountWeatherStat)
        self.app.remoteCommand.connect(self.remoteCommand)
        self.app.astrometry.signals.message.connect(self.updateAstrometryStatus)
        self.app.dome.signals.message.connect(self.updateDomeStatus)
        self.app.camera.signals.message.connect(self.updateCameraStatus)
        self.app.onlineWeather.signals.connected.connect(self.updateOnlineWeatherStat)

        # connect gui signals
        self.ui.saveConfigQuit.clicked.connect(self.quitSave)
        self.ui.loadFrom.clicked.connect(self.loadProfile)
        self.ui.saveConfigAs.clicked.connect(self.saveProfileAs)
        self.ui.saveConfig.clicked.connect(self.saveProfile)

        # connect switching of other windows
        for window in self.uiWindows:
            self.uiWindows[window]["button"].clicked.connect(self.toggleWindow)

        # initial call for writing the gui
        self.initConfig()

        # show other extended windows
        self.showExtendedWindows()

        # cyclic updates
        self.app.update1s.connect(self.updateTime)
        self.app.update1s.connect(self.updateWindowsStats)
        self.app.update1s.connect(self.smartFunctionGui)
        self.app.update1s.connect(self.smartTabGui)
        self.app.update1s.connect(self.smartEnvironGui)
        self.app.update1s.connect(self.updateWindowsStats)
        self.app.update1s.connect(self.updateDeviceStats)

    def mwSuper(self, func):
        """
        mwSuper is a replacement for super() to manage the mixin style of implementation
        it's not an ideal way to do it, but mwSuper() call the method of every ! parent
        class if they exist.

        :param func:
        :return: true for test purpose
        """

        for base in self.__class__.__bases__:
            if base.__name__ == "MWidget":
                continue

            if hasattr(base, func):
                funcAttrib = getattr(base, func)
                funcAttrib(self)

        return True

    def initConfig(self):
        """
        initConfig read the key out of the configuration dict and stores it to the gui
        elements. if some initialisations have to be proceeded with the loaded persistent
        data, they will be launched as well in this method.

        :return: True for test purpose
        """

        config = self.app.config
        self.ui.profile.setText(config.get("profileName"))
        if "mainW" not in config:
            config["mainW"] = {}

        config = config["mainW"]
        x = config.get("winPosX", 20)
        y = config.get("winPosY", 20)

        if x > self.screenSizeX:
            x = 0

        if y > self.screenSizeY:
            y = 0

        self.move(x, y)
        self.ui.mainTabWidget.setCurrentIndex(config.get("mainTabWidget", 0))
        self.ui.settingsTabWidget.setCurrentIndex(config.get("settingsTabWidget", 0))

        if not pConf.isAnalyse:
            tabWidget = self.ui.mainTabWidget.findChild(
                PyQt5.QtWidgets.QWidget, "Analyse"
            )
            tabIndex = self.ui.mainTabWidget.indexOf(tabWidget)
            self.ui.mainTabWidget.setTabEnabled(tabIndex, False)

        tabWidget = self.ui.mainTabWidget.findChild(PyQt5.QtWidgets.QWidget, "Power")
        tabIndex = self.ui.mainTabWidget.indexOf(tabWidget)
        self.ui.mainTabWidget.setTabEnabled(tabIndex, False)

        tabWidget = self.ui.mainTabWidget.findChild(PyQt5.QtWidgets.QWidget, "Relay")
        tabIndex = self.ui.mainTabWidget.indexOf(tabWidget)
        self.ui.mainTabWidget.setTabEnabled(tabIndex, False)

        self.ui.mainTabWidget.setStyleSheet(self.getStyle())

        self.mwSuper("initConfig")
        self.changeStyleDynamic(self.ui.mountConnected, "color", "gray")
        self.setupIcons()
        self.show()

        return True

    def storeConfigExtendedWindows(self):
        """

        :return: True for test purpose
        """

        config = self.app.config

        # storing extended windows data
        for window in self.uiWindows:

            # check if the window state and store it is open
            config[window] = bool(self.uiWindows[window]["classObj"])

            if config[window]:
                self.uiWindows[window]["classObj"].storeConfig()

        return True

    def storeConfig(self):
        """
        storeConfig writes the keys to the configuration dict and stores. if some
        saving has to be proceeded to persistent data, they will be launched as
        well in this method.

        :return: True for test purpose
        """

        config = self.app.config
        config["profileName"] = self.ui.profile.text()

        if "mainW" not in config:
            config["mainW"] = {}

        config = config["mainW"]
        config["winPosX"] = self.pos().x()
        config["winPosY"] = self.pos().y()
        config["mainTabWidget"] = self.ui.mainTabWidget.currentIndex()
        config["settingsTabWidget"] = self.ui.settingsTabWidget.currentIndex()

        self.mwSuper("storeConfig")
        self.storeConfigExtendedWindows()

        return True

    def closeEvent(self, closeEvent):
        """
        we overwrite the close event of the window just for the main window to close the
        application as well. because it does not make sense to have child windows open if
        main is already closed.

        :return:    nothing
        """

        self.app.timer0_1s.stop()
        self.changeStyleDynamic(self.ui.pauseModel, "pause", False)
        self.closeExtendedWindows()
        super().closeEvent(closeEvent)
        self.app.quit()

    def quitSave(self):
        """
        quitSave finished up and calls the quit save function in main for saving the parameters


        :return:    true for test purpose
        """

        self.saveProfile()
        self.app.saveConfig()
        self.close()

        return True

    def setupIcons(self):
        """
        setupIcons add icon from standard library to certain buttons for improving the
        gui of the app.

        :return:    True if success for test
        """

        # main window
        self.wIcon(self.ui.saveConfigAs, "save")
        self.wIcon(self.ui.loadFrom, "load")
        self.wIcon(self.ui.saveConfig, "save")
        self.wIcon(self.ui.saveConfigQuit, "save")
        self.wIcon(self.ui.mountOn, "power-on")
        self.wIcon(self.ui.mountOff, "power-off")
        self.wIcon(self.ui.stop, "hand")
        self.wIcon(self.ui.tracking, "target")
        self.wIcon(self.ui.followSat, "satellite")
        self.wIcon(self.ui.flipMount, "flip")
        self.wIcon(self.ui.setSiderealTracking, "sidereal")
        self.wIcon(self.ui.setLunarTracking, "lunar")
        self.wIcon(self.ui.setSolarTracking, "solar")
        self.wIcon(self.ui.park, "park")

        # model points
        self.wIcon(self.ui.loadBuildPoints, "load")
        self.wIcon(self.ui.saveBuildPoints, "save")
        self.wIcon(self.ui.saveBuildPointsAs, "save")
        self.wIcon(self.ui.clearBuildP, "trash")
        self.wIcon(self.ui.genBuildGrid, "run")
        self.wIcon(self.ui.genBuildMax, "run")
        self.wIcon(self.ui.genBuildMed, "run")
        self.wIcon(self.ui.genBuildNorm, "run")
        self.wIcon(self.ui.genBuildMin, "run")
        self.wIcon(self.ui.genBuildFile, "show")
        self.wIcon(self.ui.genBuildAlign3, "run")
        self.wIcon(self.ui.genBuildAlign6, "run")
        self.wIcon(self.ui.genBuildAlign9, "run")
        self.wIcon(self.ui.genBuildGrid, "run")
        self.wIcon(self.ui.genBuildSpiralMax, "run")
        self.wIcon(self.ui.genBuildSpiralMed, "run")
        self.wIcon(self.ui.genBuildSpiralNorm, "run")
        self.wIcon(self.ui.genBuildSpiralMin, "run")
        self.wIcon(self.ui.genBuildDSO, "run")

        # horizon
        self.wIcon(self.ui.loadHorizonMask, "load")
        self.wIcon(self.ui.saveHorizonMask, "save")
        self.wIcon(self.ui.saveHorizonMaskAs, "save")
        self.wIcon(self.ui.clearHorizonMask, "trash")

        # model
        self.wIcon(self.ui.plateSolveSync, "start")
        pixmap = QPixmap(":/pics/azimuth.png")
        self.ui.picAZ.setPixmap(pixmap)
        pixmap = QPixmap(":/pics/altitude.png")
        self.ui.picALT.setPixmap(pixmap)

        self.wIcon(self.ui.cancelModel, "cross-circle")
        self.wIcon(self.ui.runModel, "start")
        self.wIcon(self.ui.pauseModel, "pause")
        self.wIcon(self.ui.endModel, "stop_m")
        self.wIcon(self.ui.batchModel, "choose")

        # manage model
        self.wIcon(self.ui.runOptimize, "start")
        self.wIcon(self.ui.cancelOptimize, "cross-circle")
        self.wIcon(self.ui.deleteWorstPoint, "circle-minus")
        self.wIcon(self.ui.clearModel, "trash")
        self.wIcon(self.ui.openAnalyseW, "bar-chart")
        self.wIcon(self.ui.showActualModelAnalyse, "copy")
        self.wIcon(self.ui.showOriginalModelAnalyse, "copy")

        self.wIcon(self.ui.loadName, "load")
        self.wIcon(self.ui.saveName, "save")
        self.wIcon(self.ui.deleteName, "trash")
        self.wIcon(self.ui.refreshName, "reload")
        self.wIcon(self.ui.refreshModel, "reload")

        # minor planets
        self.wIcon(self.ui.progMinorPlanetsFull, "run")
        self.wIcon(self.ui.progMinorPlanetsFiltered, "run")
        self.wIcon(self.ui.progEarthRotationData, "run")
        self.wIcon(self.ui.downloadIERS, "run")

        # satellite
        self.wIcon(self.ui.stopSatelliteTracking, "cross-circle")
        self.wIcon(self.ui.startSatelliteTracking, "start")
        self.wIcon(self.ui.progSatellitesFull, "run")
        self.wIcon(self.ui.progSatellitesFiltered, "run")

        # analyse
        self.wIcon(self.ui.runFlexure, "start")
        self.wIcon(self.ui.runHysteresis, "check-circle")
        self.wIcon(self.ui.cancelAnalyse, "cross-circle")

        # tools
        self.wIcon(self.ui.renameStart, "start")
        self.wIcon(self.ui.renameInputSelect, "folder")
        self.wIcon(self.ui.posButton0, "target")
        self.wIcon(self.ui.posButton1, "target")
        self.wIcon(self.ui.posButton2, "target")
        self.wIcon(self.ui.posButton3, "target")
        self.wIcon(self.ui.posButton4, "target")
        self.wIcon(self.ui.posButton5, "target")
        self.wIcon(self.ui.posButton6, "target")
        self.wIcon(self.ui.posButton7, "target")
        self.wIcon(self.ui.posButton8, "target")
        self.wIcon(self.ui.posButton9, "target")

        self.wIcon(self.ui.moveNorth, "north")
        self.wIcon(self.ui.moveEast, "east")
        self.wIcon(self.ui.moveSouth, "south")
        self.wIcon(self.ui.moveWest, "west")
        self.wIcon(self.ui.moveNorthEast, "northEast")
        self.wIcon(self.ui.moveNorthWest, "northWest")
        self.wIcon(self.ui.moveSouthEast, "southEast")
        self.wIcon(self.ui.moveSouthWest, "southWest")
        self.wIcon(self.ui.moveNorthAltAz, "north")
        self.wIcon(self.ui.moveEastAltAz, "east")
        self.wIcon(self.ui.moveSouthAltAz, "south")
        self.wIcon(self.ui.moveWestAltAz, "west")
        self.wIcon(self.ui.moveNorthEastAltAz, "northEast")
        self.wIcon(self.ui.moveNorthWestAltAz, "northWest")
        self.wIcon(self.ui.moveSouthEastAltAz, "southEast")
        self.wIcon(self.ui.moveSouthWestAltAz, "southWest")
        self.wIcon(self.ui.stopMoveAll, "stop_m")
        self.wIcon(self.ui.moveAltAzAbsolute, "target")
        self.wIcon(self.ui.moveRaDecAbsolute, "target")

        # driver setting
        for driver in self.drivers:
            if self.drivers[driver]["uiSetup"] is not None:
                ui = self.drivers[driver]["uiSetup"]
                self.wIcon(ui, "cogs")

        self.wIcon(self.ui.ascomConnect, "link")
        self.wIcon(self.ui.ascomDisconnect, "unlink")

        # imaging
        self.wIcon(self.ui.copyFromTelescopeDriver, "copy")
        self.wIcon(self.ui.haltFocuser, "bolt-alt")
        self.wIcon(self.ui.moveFocuserIn, "exit-down")
        self.wIcon(self.ui.moveFocuserOut, "exit-up")
        self.wIcon(self.ui.coverPark, "exit-down")
        self.wIcon(self.ui.coverUnpark, "exit-up")

        # dome setting
        pixmap = QPixmap(":/dome/radius.png")
        self.ui.picDome1.setPixmap(pixmap)
        pixmap = QPixmap(":/dome/north.png")
        self.ui.picDome2.setPixmap(pixmap)
        pixmap = QPixmap(":/dome/east.png")
        self.ui.picDome3.setPixmap(pixmap)
        pixmap = QPixmap(":/dome/z_gem.png")
        self.ui.picDome4.setPixmap(pixmap)
        pixmap = QPixmap(":/dome/z_10micron.png")
        self.ui.picDome5.setPixmap(pixmap)
        pixmap = QPixmap(":/dome/gem.png")
        self.ui.picDome6.setPixmap(pixmap)
        pixmap = QPixmap(":/dome/lat.png")
        self.ui.picDome7.setPixmap(pixmap)
        pixmap = QPixmap(":/dome/shutter.png")
        self.ui.picDome8.setPixmap(pixmap)
        self.wIcon(self.ui.copyFromDomeDriver, "copy")
        self.wIcon(self.ui.domeCloseShutter, "exit-down")
        self.wIcon(self.ui.domeOpenShutter, "exit-up")
        self.wIcon(self.ui.domeAbortSlew, "bolt-alt")

        # park positions
        self.wIcon(self.ui.posSave0, "download")
        self.wIcon(self.ui.posSave1, "download")
        self.wIcon(self.ui.posSave2, "download")
        self.wIcon(self.ui.posSave3, "download")
        self.wIcon(self.ui.posSave4, "download")
        self.wIcon(self.ui.posSave5, "download")
        self.wIcon(self.ui.posSave6, "download")
        self.wIcon(self.ui.posSave7, "download")
        self.wIcon(self.ui.posSave8, "download")
        self.wIcon(self.ui.posSave9, "download")

        # misc setting
        self.wIcon(self.ui.installVersion, "world")

        return True

    def updateMountConnStat(self, status):
        """
        updateMountConnStat show the connection status of the mount. if status is None,
        which means there is no valid host entry for connection, the status is grey

        :param status:
        :return: true for test purpose
        """
        self.deviceStat["mount"] = status
        return True

    def updateMountWeatherStat(self, setting):
        """
        updateMountWeatherStat show the connection status of the mount weather station
        connected. if the data values are None there is no station attached to the
        GPS port,

        :return: true for test purpose
        """
        if setting.weatherTemperature is None and setting.weatherPressure is None:
            self.deviceStat["directWeather"] = None

        else:
            if setting.weatherStatus is None:
                self.deviceStat["directWeather"] = False

            else:
                self.deviceStat["directWeather"] = True
        return True

    def smartFunctionGui(self):
        """
        smartFunctionGui enables and disables gui actions depending on the actual state of the
        different devices. this should be the core of avoiding user misused during running
        operations. smartGui is run every 1 second synchronously, because it can't be
        simpler done with dynamic approach. all different situations in a running
        environment is done locally.

        :return: true for test purpose
        """
        isModelingReady = all(
            self.deviceStat[x] for x in ["mount", "camera", "astrometry"]
        )

        if isModelingReady and self.app.data.buildP:
            self.ui.runModel.setEnabled(True)
            self.ui.plateSolveSync.setEnabled(True)
            self.ui.runFlexure.setEnabled(True)
            self.ui.runHysteresis.setEnabled(True)

        else:
            self.ui.runModel.setEnabled(False)
            self.ui.plateSolveSync.setEnabled(False)
            self.ui.runFlexure.setEnabled(False)
            self.ui.runHysteresis.setEnabled(False)

        if self.deviceStat.get("mount", False):
            self.ui.batchModel.setEnabled(True)

        else:
            self.ui.batchModel.setEnabled(False)

        stat = self.deviceStat.get("environOverall", None)

        if stat is None:
            self.ui.refractionGroup.setEnabled(False)
            self.ui.setRefractionManual.setEnabled(False)

        elif stat and self.deviceStat.get("mount", None):
            self.ui.refractionGroup.setEnabled(True)
            self.ui.setRefractionManual.setEnabled(True)

        else:
            self.ui.refractionGroup.setEnabled(False)
            self.ui.setRefractionManual.setEnabled(False)
        return True

    def smartTabGui(self):
        """
        smartTabGui enables and disables tab visibility depending on the actual state of the
        different devices.
        :return: true for test purpose
        """
        smartTabs = {
            "Power": {
                "statID": "power",
                "tab": self.ui.mainTabWidget,
            },
            "Relay": {
                "statID": "relay",
                "tab": self.ui.mainTabWidget,
            },
            # 'RelayS': {'statID': 'relay',
            #            'tab': self.ui.settingsTabWidget,
            #            },
        }
        tabChanged = False

        for key, tab in smartTabs.items():
            # finding the right tab and get the status
            tabWidget = smartTabs[key]["tab"].findChild(PyQt5.QtWidgets.QWidget, key)
            tabIndex = smartTabs[key]["tab"].indexOf(tabWidget)
            tabStatus = smartTabs[key]["tab"].isTabEnabled(tabIndex)

            # determine the new stat, set it and check if changed
            stat = bool(self.deviceStat.get(smartTabs[key]["statID"]))
            smartTabs[key]["tab"].setTabEnabled(tabIndex, stat)
            tabChanged = tabChanged or (tabStatus != stat)

        # redraw tabs only when a change occurred. this is necessary, because
        # enable and disable does not remove tabs
        if tabChanged:
            self.ui.mainTabWidget.setStyleSheet(self.getStyle())
            self.ui.settingsTabWidget.setStyleSheet(self.getStyle())
        return True

    def smartEnvironGui(self):
        """
        smartEnvironGui enables and disables gui actions depending on the actual state
        of the different environment devices. it is run every 1 second synchronously,
        because it can't be simpler done with dynamic approach. all different situations
        in a running environment is done locally.

        :return: true for test purpose
        """
        environ = {
            "directWeather": self.ui.directWeatherGroup,
            "sensorWeather": self.ui.sensorWeatherGroup,
            "onlineWeather": self.ui.onlineWeatherGroup,
            "skymeter": self.ui.skymeterGroup,
            "powerWeather": self.ui.powerGroup,
        }

        for key, group in environ.items():
            stat = self.deviceStat.get(key, None)

            if stat is None:
                group.setFixedWidth(0)
                group.setEnabled(False)

            elif stat:
                group.setMinimumSize(75, 0)
                group.setEnabled(True)

            else:
                group.setMinimumSize(75, 0)
                group.setEnabled(False)
        return True

    def updateWindowsStats(self):
        """
        :return: True for test purpose
        """
        for win in self.app.uiWindows:
            winObj = self.app.uiWindows[win]

            if winObj["classObj"]:
                self.changeStyleDynamic(winObj["button"], "running", True)

            else:
                self.changeStyleDynamic(winObj["button"], "running", False)
        return True

    def updateDeviceStats(self):
        """
        updateDeviceStats sets the colors in main window upper bar for getting
        important overview, which functions are available.

        the refraction sources etc are defined in tabEnviron, but it is optimal
        setting the selected source right at this point as it is synchronous if
        state is switching

        :return: True for test purpose
        """
        if self.refractionSource in self.deviceStat:
            self.deviceStat["environOverall"] = self.deviceStat[self.refractionSource]

        else:
            self.deviceStat["environOverall"] = None

        for device, ui in self.deviceStatGui.items():
            if self.deviceStat.get(device, None) is None:
                self.changeStyleDynamic(ui, "color", "gray")

            elif self.deviceStat[device]:
                self.changeStyleDynamic(ui, "color", "green")

            else:
                self.changeStyleDynamic(ui, "color", "red")
        return True

    def updateOnlineWeatherStat(self, stat):
        """
        :param stat:
        :return: True for test purpose
        """
        self.deviceStat["onlineWeather"] = stat
        return True

    def updateTime(self):
        """
        :return: True for test purpose
        """
        self.ui.timeComputer.setText(datetime.now().strftime("%H:%M:%S"))
        if self.ui.isOnline.isChecked():
            text = "Internet Online Mode"

        else:
            text = "Offline Mode"

        text = f"{self.threadPool.activeThreadCount():2d} - {text}"
        self.ui.statusOnline.setTitle(text)
        return True

    def updateAstrometryStatus(self, text):
        """
        :param text:
        :return: true for test purpose
        """
        self.ui.astrometryText.setText(text)
        return True

    def updateDomeStatus(self, text):
        """
        :param text:
        :return: true for test purpose
        """
        self.ui.domeText.setText(text)
        return True

    def updateCameraStatus(self, text):
        """
        :param text:
        :return: true for test purpose
        """
        self.ui.cameraText.setText(text)
        return True

    def updateStatusGUI(self, obs):
        """
        :return:    True if ok for testing
        """
        if obs.statusText() is not None:
            self.ui.statusText.setText(obs.statusText())

        else:
            self.ui.statusText.setText("-")

        if self.app.mount.obsSite.status == 0:
            self.changeStyleDynamic(self.ui.tracking, "running", "true")
        else:
            self.changeStyleDynamic(self.ui.tracking, "running", "false")

        if self.app.mount.obsSite.status == 5:
            self.changeStyleDynamic(self.ui.park, "running", "true")
        else:
            self.changeStyleDynamic(self.ui.park, "running", "false")

        if self.app.mount.obsSite.status == 1:
            self.changeStyleDynamic(self.ui.stop, "running", "true")
        else:
            self.changeStyleDynamic(self.ui.stop, "running", "false")

        return True

    def deleteWindowResource(self, widget=None):
        """

        :return: success
        """

        if not widget:
            return False

        for window in self.uiWindows:
            if self.uiWindows[window]["name"] != widget.objectName():
                continue

            self.uiWindows[window]["classObj"] = None
        gc.collect()

        return True

    def buildWindow(self, window):
        """
        buildWindow makes new object instance from window class. both are stored in the
        uiWindows dict for usage.

        :return: true for test purpose
        """
        self.uiWindows[window]["classObj"] = self.uiWindows[window]["class"](self.app)
        self.uiWindows[window]["classObj"].destroyed.connect(self.deleteWindowResource)
        self.uiWindows[window]["classObj"].initConfig()
        self.uiWindows[window]["classObj"].showWindow()
        return True

    def toggleWindow(self):
        """
        :return: true for test purpose
        """
        for window in self.uiWindows:
            if self.uiWindows[window]["button"] != self.sender():
                continue

            if not self.uiWindows[window]["classObj"]:
                self.buildWindow(window)

            else:
                self.uiWindows[window]["classObj"].close()
        return True

    def showExtendedWindows(self):
        """
        showExtendedWindows opens all extended windows depending on their opening status
        stored in the configuration dict.
        :return: true for test purpose
        """
        for window in self.uiWindows:
            if not self.app.config.get(window, False):
                continue
            if window == "showSimulatorW":
                continue

            self.buildWindow(window)
        return True

    def closeExtendedWindows(self):
        """
        closeExtendedWindows closes all open extended windows by calling close and
        waits until the window class is deleted.

        :return: true for test purpose
        """

        for window in self.uiWindows:
            if not self.uiWindows[window]["classObj"]:
                continue

            self.uiWindows[window]["classObj"].close()

        waitDeleted = True
        while waitDeleted:
            for window in self.uiWindows:
                if self.uiWindows[window]["classObj"]:
                    continue

                waitDeleted = False

            QTest.qWait(100)

        return True

    @staticmethod
    def checkExtension(filePath, ext):
        """
        checkExtension ensures to have an extension attached for a filename.

        :param filePath:
        :param ext:
        :return:
        """

        if not filePath.endswith(ext):
            filePath += ext

        return filePath

    def loadProfile(self):
        """
        loadProfile interacts to get a new profile name. if a valid is received, it closes
        all extended windows to be sure to have all setup saved, than load the new profile
        and initializes all classes and opens the necessary extended windows with their
        setups stored.

        loadProfile does not save the actual configuration before loading another one.

        :return:
        """

        folder = self.app.mwGlob["configDir"]
        loadFilePath, name, ext = self.openFile(
            self,
            "Open config file",
            folder,
            "Config files (*.cfg)",
            enableDir=False,
        )
        if not name:
            return False

        # closing all windows to be base lined
        self.closeExtendedWindows()

        suc = self.app.loadConfig(name=name)

        if suc:
            self.app.config["profileName"] = name
            self.ui.profile.setText(name)
            self.app.message.emit(f"Profile              [{name}] loaded", 0)

        else:
            self.app.message.emit(
                f"Profile              [{name}] cannot no be loaded", 2
            )

        # configure mainApp
        topo = self.app.initConfig()
        self.app.mount.obsSite.location = topo

        # initialize the mainW configurations
        self.initConfig()

        # instantiate the extended windows where needed
        self.showExtendedWindows()

        return True

    def saveProfileAs(self):
        """

        :return:
        """

        folder = self.app.mwGlob["configDir"]
        saveFilePath, name, ext = self.saveFile(
            self,
            "Save config file",
            folder,
            "Config files (*.cfg)",
            enableDir=False,
        )
        if not name:
            return False

        self.storeConfig()
        self.app.storeConfig()
        suc = self.app.saveConfig(name=name)
        if suc:
            self.ui.profile.setText(name)
            self.app.message.emit(f"Profile              [{name}] saved", 0)

        else:
            self.app.message.emit(
                f"Profile              [{name}] cannot no be saved", 2
            )

        return True

    def saveProfile(self):
        """
        saveProfile calls save profile in main and sends a message to the user about
        success.

        :return: nothing
        """

        # self.app.config = {}
        self.storeConfig()
        self.app.storeConfig()

        suc = self.app.saveConfig(name=self.ui.profile.text())
        if suc:
            self.app.message.emit("Actual profile saved", 0)

        else:
            self.app.message.emit("Actual profile cannot not be saved", 2)

        return suc

    def remoteCommand(self, command):
        """
        remoteCommand received signals from remote class and executes them.

        :param command:
        :return: True for test purpose
        """

        if command == "shutdown":
            self.quitSave()
            self.app.message.emit("Shutdown MW remotely", 2)

        elif command == "shutdown mount":
            self.mountShutdown()
            self.app.message.emit("Shutdown mount remotely", 2)

        elif command == "boot mount":
            self.mountBoot()
            self.app.message.emit("Boot mount remotely", 2)

        return True



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
#
# written in python3, (c) 2019-2021 by mworion
# Licence APL2.0
#
###########################################################
# standard libraries

# external packages
import numpy as np
from PyQt5.QtGui import QVector3D
from PyQt5.Qt3DExtras import QCuboidMesh
from PyQt5.Qt3DExtras import QCylinderMesh
from PyQt5.Qt3DCore import QEntity, QTransform
from skyfield import functions

# local import
from gui.extWindows.simulator.materials import Materials


class SimulatorHorizon:

    __all__ = ['SimulatorHorizon',
               ]

    def __init__(self, app):
        self.app = app
        self.horizon = []
        self.horizonRoot = None

    @staticmethod
    def createLine(rEntity, dx, dy, dz):
        """
        create line draw a line between two point or better along dx, dy, dz.
        Therefore three transformations are made and the resulting vector has to
        be translated half the length, because is will be drawn symmetrically to
        the starting point.

        :param rEntity:
        :param dx:
        :param dy:
        :param dz:
        :return:
        """
        radius, alt, az = functions.to_spherical(np.array((dx, dy, dz)))
        az = np.degrees(az)
        alt = np.degrees(alt)

        e1 = QEntity(rEntity)
        trans1 = QTransform()
        trans1.setRotationZ(az + 90)
        e1.addComponent(trans1)

        e2 = QEntity(e1)
        trans2 = QTransform()
        trans2.setRotationX(-alt)
        e2.addComponent(trans2)

        e3 = QEntity(e2)
        mesh = QCylinderMesh()
        mesh.setRadius(0.008)
        mesh.setLength(radius)
        trans3 = QTransform()
        trans3.setTranslation(QVector3D(0, radius / 2, 0))
        e3.addComponent(mesh)
        e3.addComponent(trans3)
        e3.addComponent(Materials().lines)
        return e3

    @staticmethod
    def createWall(rEntity, alt, az, space):
        """
        createWall draw a plane in radius distance to show the horizon. spacing
        is the angular spacing between this planes

        :param rEntity:
        :param alt:
        :param az:
        :param space:
        :return: entity
        """
        radius = 4
        e1 = QEntity(rEntity)
        trans1 = QTransform()
        trans1.setRotationZ(-az)
        e1.addComponent(trans1)

        e2 = QEntity(e1)
        trans2 = QTransform()
        trans2.setTranslation(QVector3D(radius, 0, 0))
        e2.addComponent(trans2)

        e3 = QEntity(e2)
        height = radius * np.tan(np.radians(alt)) + 1.35
        mesh = QCuboidMesh()
        mesh.setXExtent(0.01)
        mesh.setYExtent(radius * np.tan(np.radians(space)))
        mesh.setZExtent(height)
        trans3 = QTransform()
        trans3.setTranslation(QVector3D(0, 0, height / 2))
        e3.addComponent(mesh)
        e3. addComponent(trans3)
        e3.addComponent(Materials().walls)
        return e3

    def create(self, rEntity, show):
        """
        createHorizon draws a horizon "wall" by circling over the horizon points
        and putting cuboid meshed around a circle with defined radius. the space
        is the angle width of a plane in degrees

        :return: success
        """
        if self.horizon:
            self.horizonRoot.setParent(None)

        self.horizon.clear()

        if not show:
            return False

        if not self.app.data.horizonP:
            return False

        self.horizonRoot = QEntity(rEntity)

        space = 5
        horizonAz = np.linspace(0, 360 - space, int(360 / space))
        alt = [x[0] for x in self.app.data.horizonP]
        az = [x[1] for x in self.app.data.horizonP]
        horizonAlt = np.interp(horizonAz, az, alt)

        for alt, az in zip(horizonAlt, horizonAz):
            e = self.createWall(self.horizonRoot, alt, az, space)
            element = {'e': e}
            self.horizon.append(element)
        return True



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
#
# written in python3, (c) 2019-2021 by mworion
#
# Licence APL2.0
#
###########################################################
# standard libraries
import logging
from threading import Lock
import bisect

# external packages
import numpy as np
from PyQt5.QtWidgets import QVBoxLayout
from matplotlib.figure import Figure
from matplotlib.ticker import FormatStrFormatter
from matplotlib.backends.backend_qt5agg import FigureCanvas
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg

# local imports

__all__ = [
    'ToolsMatplotlib',
]


class ToolsMatplotlib:
    """
    """

    __all__ = ['ToolsMatplotlib',
               ]

    log = logging.getLogger(__name__)

    @staticmethod
    def findIndexValue(ui, searchString, relaxed=False):
        """
        :param ui:
        :param searchString:
        :param relaxed:
        :return:
        """
        for index in range(ui.model().rowCount()):
            modelIndex = ui.model().index(index, 0)
            indexValue = ui.model().data(modelIndex)

            if not indexValue:
                continue

            if relaxed:
                if searchString in indexValue:
                    return index

            else:
                if indexValue.startswith(searchString):
                    return index

        return 0

    @staticmethod
    def embedMatplot(widget=None, constrainedLayout=True):
        """
        embedMatplot provides the wrapper to use matplotlib drawings inside a pyqt5
        application gui. you call it with the parent widget, which is linked to matplotlib
        canvas of the same size. the background is set to transparent, so you could layer
        multiple figures on top.

        :param      widget:         parent ui element, which is the reference for embedding
        :param      constrainedLayout:
        :return:    staticCanvas:   matplotlib reference as parent for figures
        """

        if not widget:
            return None

        widget.setStyleSheet("background:transparent;")
        layout = QVBoxLayout(widget)
        layout.setContentsMargins(0, 0, 0, 0)
        staticCanvas = FigureCanvas(Figure(dpi=75,
                                           facecolor='None',
                                           frameon=False,
                                           tight_layout=constrainedLayout,
                                           )
                                    )
        FigureCanvasQTAgg.updateGeometry(staticCanvas)
        layout.addWidget(staticCanvas)
        staticCanvas.setParent(widget)

        return staticCanvas

    def generatePolar(self, widget=None, title='', horizon=False, showAxes=True):
        """

        :param widget:
        :param title:
        :param horizon:
        :param showAxes:
        :return:
        """

        if widget is None:
            return None, None

        if not hasattr(widget, 'figure'):
            return None, None

        lock = Lock()

        if showAxes:
            color = self.M_BLUE
            colorGrid = self.M_GREY

        else:
            color = self.M_TRANS
            colorGrid = self.M_TRANS

        with lock:
            figure = widget.figure

            if figure.axes:
                axe = figure.axes[0]
                axe.cla()

            else:
                figure.clf()
                axe = figure.add_subplot(1, 1, 1, polar=True, facecolor='None')

            axe.grid(True, color=colorGrid)

            if title:
                axe.set_title(title, color=color, fontweight='bold', pad=15)

            axe.set_xlabel('', color=color, fontweight='bold', fontsize=12)
            axe.set_ylabel('', color=color, fontweight='bold', fontsize=12)
            axe.tick_params(axis='x', colors=color, labelsize=12)
            axe.tick_params(axis='y', colors=color, labelsize=12)
            axe.set_theta_zero_location('N')
            axe.set_rlabel_position(45)
            axe.set_theta_direction(-1)
            axe.spines['polar'].set_color(color)

            # ticks have to be set before labels to be sure to have them positioned correctly
            axe.set_xticks(np.radians([0, 45, 90, 135, 180, 225, 270, 315]))
            axe.set_xticklabels(['N', 'NE', 'E', 'SE', 'S', 'SW', 'W', 'NW'])

            if not horizon:
                return axe, figure

            axe.set_ylim(0, 90)
            axe.set_yticks(range(0, 91, 15))
            axe.set_yticklabels(['', '75', '60', '45', '30', '15', ''])

            return axe, figure

    def generateFlat(self, widget=None, title='', horizon=False, showAxes=True):
        """

        :param widget:
        :param title:
        :param horizon:
        :param showAxes:
        :return:
        """

        if widget is None:
            return None, None

        if not hasattr(widget, 'figure'):
            return None, None

        lock = Lock()

        if showAxes:
            color = self.M_BLUE

        else:
            color = self.M_TRANS

        with lock:
            figure = widget.figure

            if figure.axes:
                axe = figure.axes[0]
                axe.cla()

            else:
                figure.clf()
                axe = figure.add_subplot(1, 1, 1, facecolor='None')

            axe.spines['bottom'].set_color(color)
            axe.spines['top'].set_color(color)
            axe.spines['left'].set_color(color)
            axe.spines['right'].set_color(color)

            if showAxes:
                axe.grid(showAxes, color=self.M_GREY)

            else:
                axe.grid(showAxes)

            if title:
                axe.set_title(title, color=color, fontweight='bold', pad=15)

            axe.set_xlabel('', color=color, fontweight='bold', fontsize=12)
            axe.set_ylabel('', color=color, fontweight='bold', fontsize=12)
            axe.tick_params(axis='x', colors=color, labelsize=12)
            axe.tick_params(axis='y', colors=color, labelsize=12)

            if not horizon:
                return axe, figure

            axe.set_xlim(0, 360)
            axe.set_ylim(0, 90)

            axe.set_xticks(np.arange(0, 361, 45))
            axe.set_xticklabels(['0 N', '45 NE', '90 E', '135 SE', '180 S', '225 SW', '270 W',
                                 '315 NW', '360 N'])
            axe.set_xlabel('Azimuth [degrees]', color=color, fontweight='bold', fontsize=12)
            axe.set_ylabel('Altitude [degrees]', color=color, fontweight='bold', fontsize=12)

            return axe, figure

    def generateColorbar(self, figure=None, scatter=None, label=''):
        """
        :param figure:
        :param scatter:
        :param label:
        :return:
        """
        if len(figure.axes) > 1:
            return False

        formatString = FormatStrFormatter('%1.0f')
        figure.colorbar(scatter,
                        pad=0.1,
                        fraction=0.12,
                        aspect=25,
                        shrink=0.9,
                        format=formatString,
                        )
        figure.axes[1].set_ylabel(label, color=self.M_BLUE)
        figure.axes[1].tick_params(axis='y', labelcolor=self.M_BLUE, color=self.M_BLUE)

        return True

    @staticmethod
    def getIndexPoint(event=None, plane=None, epsilon=2):
        """
        getIndexPoint returns the index of the point which is nearest to the coordinate
        of the mouse click when the click is in distance epsilon of the points. otherwise
        no index will be returned.

        :param event: data of the mouse clicked event
        :param plane: coordinates as tuples (alt, az)
        :param epsilon:
        :return: index or none
        """
        if event is None:
            return None

        if plane is None:
            return None

        if len(plane) == 0:
            return 0

        xt = np.asarray([i[1] for i in plane])
        yt = np.asarray([i[0] for i in plane])
        d = np.sqrt((xt - event.xdata)**2 / 4 + (yt - event.ydata)**2)
        index = d.argsort()[:1][0]

        if d[index] >= epsilon:
            return None

        index = int(index)
        return index

    @staticmethod
    def getIndexPointX(x=None, plane=None):
        """
        getIndexPointX returns the index of the point which has a x coordinate closest to
        the left of the x coordinate of the mouse click regardless which y coordinate it has

        :param x: data of the mouse clicked event
        :param plane: coordinates as tuples (x, y)
        :return: index or none
        """
        if x is None:
            return None

        if not plane:
            return None

        xt = [i[1] for i in plane]
        index = int(bisect.bisect_left(xt, x))
        return index

    @staticmethod
    def writeRetrofitData(mountModel, buildModel):
        """

        :param mountModel:
        :param buildModel:
        :return:
        """

        for i, mPoint in enumerate(buildModel):
            mPoint['errorRMS'] = mountModel.starList[i].errorRMS
            mPoint['errorAngle'] = mountModel.starList[i].errorAngle.degrees
            mPoint['haMountModel'] = mountModel.starList[i].coord.ra.hours
            mPoint['decMountModel'] = mountModel.starList[i].coord.dec.degrees
            mPoint['errorRA'] = mountModel.starList[i].errorRA()
            mPoint['errorDEC'] = mountModel.starList[i].errorDEC()
            mPoint['errorIndex'] = mountModel.starList[i].number
            mPoint['modelTerms'] = mountModel.terms
            mPoint['modelErrorRMS'] = mountModel.errorRMS
            mPoint['modelOrthoError'] = mountModel.orthoError.degrees * 3600
            mPoint['modelPolarError'] = mountModel.polarError.degrees * 3600

        return buildModel



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
#
# written in python3, (c) 2019-2021 by mworion
# Licence APL2.0
#
###########################################################
# standard libraries

# external packages

# local import


class SettDome(object):
    """
    """

    def __init__(self):
        self.ui.domeRadius.valueChanged.connect(self.setUseGeometryInMount)
        self.ui.offGEM.valueChanged.connect(self.setUseGeometryInMount)
        self.ui.offLAT.valueChanged.connect(self.setUseGeometryInMount)
        self.ui.domeEastOffset.valueChanged.connect(self.setUseGeometryInMount)
        self.ui.domeNorthOffset.valueChanged.connect(self.setUseGeometryInMount)
        self.ui.domeZoffGEM.valueChanged.connect(self.setZoffGEMInMount)
        self.ui.domeZoff10micron.valueChanged.connect(self.setZoff10micronInMount)
        self.ui.domeShutterWidth.valueChanged.connect(self.setUseGeometryInMount)
        self.ui.settleTimeDome.valueChanged.connect(self.setDomeSettlingTime)
        self.ui.checkUseDomeGeometry.clicked.connect(self.setUseDomeGeometry)
        self.ui.copyFromDomeDriver.clicked.connect(self.updateDomeGeometryToGui)
        self.app.mount.signals.firmwareDone.connect(self.setUseGeometryInMount)
        self.app.mount.signals.firmwareDone.connect(self.setZoffGEMInMount)

        self.ui.domeRadius.valueChanged.connect(self.tab1)
        self.ui.domeNorthOffset.valueChanged.connect(self.tab2)
        self.ui.domeEastOffset.valueChanged.connect(self.tab3)
        self.ui.domeZoffGEM.valueChanged.connect(self.tab4)
        self.ui.domeZoff10micron.valueChanged.connect(self.tab5)
        self.ui.offGEM.valueChanged.connect(self.tab6)
        self.ui.offLAT.valueChanged.connect(self.tab7)
        self.ui.domeShutterWidth.valueChanged.connect(self.tab8)
        self.app.update1s.connect(self.updateShutterStatGui)
        self.ui.domeAbortSlew.clicked.connect(self.domeAbortSlew)
        self.ui.domeOpenShutter.clicked.connect(self.domeOpenShutter)
        self.ui.domeCloseShutter.clicked.connect(self.domeCloseShutter)

    def tab1(self):
        self.ui.tabDomeExplain.setCurrentIndex(0)
        self.ui.tabDomeExplain.setStyleSheet(self.getStyle())

    def tab2(self):
        self.ui.tabDomeExplain.setCurrentIndex(1)
        self.ui.tabDomeExplain.setStyleSheet(self.getStyle())

    def tab3(self):
        self.ui.tabDomeExplain.setCurrentIndex(2)
        self.ui.tabDomeExplain.setStyleSheet(self.getStyle())

    def tab4(self):
        self.ui.tabDomeExplain.setCurrentIndex(3)
        self.ui.tabDomeExplain.setStyleSheet(self.getStyle())

    def tab5(self):
        self.ui.tabDomeExplain.setCurrentIndex(4)
        self.ui.tabDomeExplain.setStyleSheet(self.getStyle())

    def tab6(self):
        self.ui.tabDomeExplain.setCurrentIndex(5)
        self.ui.tabDomeExplain.setStyleSheet(self.getStyle())

    def tab7(self):
        self.ui.tabDomeExplain.setCurrentIndex(6)
        self.ui.tabDomeExplain.setStyleSheet(self.getStyle())

    def tab8(self):
        self.ui.tabDomeExplain.setCurrentIndex(7)
        self.ui.tabDomeExplain.setStyleSheet(self.getStyle())

    def initConfig(self):
        """
        initConfig read the key out of the configuration dict and stores it to the gui
        elements. if some initialisations have to be proceeded with the loaded persistent
        data, they will be launched as well in this method.

        :return: True for test purpose
        """
        config = self.app.config['mainW']
        self.ui.domeShutterWidth.setValue(config.get('domeShutterWidth', 0.2))
        self.ui.domeNorthOffset.setValue(config.get('domeNorthOffset', 0))
        self.ui.domeEastOffset.setValue(config.get('domeEastOffset', 0))
        self.ui.domeZoffGEM.setValue(config.get('domeZoffGEM', 0))
        self.ui.offGEM.setValue(config.get('offGEM', 0))
        self.ui.offLAT.setValue(config.get('offLAT', 0))
        self.ui.domeRadius.setValue(config.get('domeRadius', 1.5))
        self.ui.checkUseDomeGeometry.setChecked(config.get('checkUseDomeGeometry', False))
        self.ui.checkAutomaticDome.setChecked(config.get('checkAutomaticDome', False))
        self.ui.settleTimeDome.setValue(config.get('settleTimeDome', 0))
        self.setUseDomeGeometry()
        return True

    def storeConfig(self):
        """
        storeConfig writes the keys to the configuration dict and stores. if some
        saving has to be proceeded to persistent data, they will be launched as
        well in this method.

        :return: True for test purpose
        """
        config = self.app.config['mainW']
        config['domeRadius'] = self.ui.domeRadius.value()
        config['domeShutterWidth'] = self.ui.domeShutterWidth.value()
        config['domeNorthOffset'] = self.ui.domeNorthOffset.value()
        config['domeEastOffset'] = self.ui.domeEastOffset.value()
        config['domeZoffGEM'] = self.ui.domeZoffGEM.value()
        config['offGEM'] = self.ui.offGEM.value()
        config['offLAT'] = self.ui.offLAT.value()
        config['checkUseDomeGeometry'] = self.ui.checkUseDomeGeometry.isChecked()
        config['checkAutomaticDome'] = self.ui.checkAutomaticDome.isChecked()
        config['settleTimeDome'] = self.ui.settleTimeDome.value()
        return True

    def setZoffGEMInMount(self):
        """
        :return:
        """
        self.app.mount.geometry.offVertGEM = self.ui.domeZoffGEM.value()
        self.ui.domeZoff10micron.setValue(self.app.mount.geometry.offVert)
        self.app.updateDomeSettings.emit()
        return True

    def setZoff10micronInMount(self):
        """
        :return:
        """
        self.app.mount.geometry.offVert = self.ui.domeZoff10micron.value()
        self.ui.domeZoffGEM.setValue(self.app.mount.geometry.offVertGEM)
        self.app.updateDomeSettings.emit()
        return True

    def setUseGeometryInMount(self):
        """
        setUseGeometryInMount updates the mount class with the new setting if use
        geometry for dome calculation should be used or not.

        :return: true for test purpose
        """
        if self.ui.checkAutomaticDome.isChecked():
            self.updateDomeGeometryToGui()

        self.app.mount.geometry.domeRadius = self.ui.domeRadius.value()
        self.app.mount.geometry.offGEM = self.ui.offGEM.value()
        self.app.mount.geometry.offLAT = self.ui.offLAT.value()
        self.app.mount.geometry.offNorth = self.ui.domeNorthOffset.value()
        self.app.mount.geometry.offEast = self.ui.domeEastOffset.value()
        self.app.dome.domeShutterWidth = self.ui.domeShutterWidth.value()
        self.app.updateDomeSettings.emit()
        return True

    def setUseDomeGeometry(self):
        """
        :return: True for test purpose
        """
        useGeometry = self.ui.checkUseDomeGeometry.isChecked()
        self.app.dome.useGeometry = useGeometry
        return True

    def updateDomeGeometryToGui(self):
        """
        :return: true for test purpose
        """
        value = float(self.app.dome.data.get('DOME_MEASUREMENTS.DM_OTA_OFFSET', 0))
        self.ui.offGEM.setValue(value)

        value = float(self.app.dome.data.get('DOME_MEASUREMENTS.DM_DOME_RADIUS', 0))
        self.ui.domeRadius.setValue(value)

        value = float(self.app.dome.data.get('DOME_MEASUREMENTS.DM_SHUTTER_WIDTH', 0))
        self.ui.domeShutterWidth.setValue(value)

        value = float(self.app.dome.data.get('DOME_MEASUREMENTS.DM_NORTH_DISPLACEMENT', 0))
        self.ui.domeNorthOffset.setValue(value)

        value = float(self.app.dome.data.get('DOME_MEASUREMENTS.DM_EAST_DISPLACEMENT', 0))
        self.ui.domeEastOffset.setValue(value)

        value = float(self.app.dome.data.get('DOME_MEASUREMENTS.DM_UP_DISPLACEMENT', 0))
        self.ui.domeZoffGEM.setValue(value)
        return True

    def setDomeSettlingTime(self):
        """
        :return: true for test purpose
        """
        self.app.dome.settlingTime = self.ui.settleTimeDome.value()
        return True

    def updateShutterStatGui(self):
        """
        :return: True for test purpose
        """
        value = self.app.dome.data.get('DOME_SHUTTER.SHUTTER_OPEN', None)
        if value is True:
            self.changeStyleDynamic(self.ui.domeOpenShutter, 'running', True)
            self.changeStyleDynamic(self.ui.domeCloseShutter, 'running', False)
        elif value is False:
            self.changeStyleDynamic(self.ui.domeOpenShutter, 'running', False)
            self.changeStyleDynamic(self.ui.domeCloseShutter, 'running', True)
        else:
            self.changeStyleDynamic(self.ui.domeOpenShutter, 'running', False)
            self.changeStyleDynamic(self.ui.domeCloseShutter, 'running', False)

        value = self.app.dome.data.get('Status.Shutter', None)
        if value:
            self.ui.domeShutterStatusText.setText(value)
        return True

    def domeAbortSlew(self):
        """
        :return:
        """
        suc = self.app.dome.abortSlew()
        if not suc:
            self.app.message.emit('Dome slew abort could not be executed', 2)

        return suc

    def domeOpenShutter(self):
        """
        :return:
        """
        suc = self.app.dome.openShutter()
        if not suc:
            self.app.message.emit('Dome open shutter could not be executed', 2)

        return suc

    def domeCloseShutter(self):
        """
        :return:
        """
        suc = self.app.dome.closeShutter()
        if not suc:
            self.app.message.emit('Dome close shutter could not be executed', 2)

        return suc



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
#
# written in python3, (c) 2019-2021 by mworion
#
# Licence APL2.0
#
###########################################################
# standard libraries
import logging
import re
import time

# external packages
from PyQt5.QtCore import QObject
from PyQt5.QtCore import pyqtSignal
from PyQt5.QtCore import QMutex
from PyQt5.QtCore import QTimer
import requests

# local imports


class RelaySignals(QObject):
    """
    The RelaySignals class offers a list of signals to be used and instantiated by
    the Relay class to get signals for triggers for finished tasks to
    enable a gui to update their values transferred to the caller back.

    This has to be done in a separate class as the signals have to be subclassed from
    QObject and the Relay class itself is subclassed from object
    """

    __all__ = ['RelaySignals']

    statusReady = pyqtSignal()

    serverConnected = pyqtSignal()
    serverDisconnected = pyqtSignal(object)
    deviceConnected = pyqtSignal(str)
    deviceDisconnected = pyqtSignal(str)


class KMRelay(QObject):
    """
    The class KMRelay inherits all information and handling of KMtronic relay board
    attributes of the connected board and provides the abstracted interface.

        >>> relay = KMRelay()

    """

    __all__ = ['KMRelay',
               ]

    log = logging.getLogger(__name__)

    # polling cycle for relay box
    CYCLE_POLLING = 1000
    # default port for KMTronic Relay
    DEFAULT_PORT = 80
    # timeout for requests
    TIMEOUT = 0.5
    # width for pulse
    PULSEWIDTH = 0.5

    def __init__(self):
        super().__init__()

        self.signals = RelaySignals()
        # minimum set for driver package built in
        self.framework = ''
        self.data = {}
        self.defaultConfig = {
            'framework': '',
            'frameworks': {
                'relay': {
                    'deviceName': 'KMRelay',
                    'hostaddress': '',
                    'user': '',
                    'password': '',
                }
            }
        }
        self.run = {
            'relay': self
        }

        self.mutexPoll = QMutex()
        self.deviceName = ''
        self.hostaddress = ''
        self.user = ''
        self.password = ''
        self.status = [0] * 8

        self.timerTask = QTimer()
        self.timerTask.setSingleShot(False)
        self.timerTask.timeout.connect(self.cyclePolling)

    def startCommunication(self, loadConfig=False):
        """
        startCommunication enables the cyclic timers for polling necessary relay data.

        :param loadConfig:
        :return: success
        """
        if not self.hostaddress:
            return False

        self.timerTask.start(self.CYCLE_POLLING)
        # self.signals.deviceConnected.emit('KMTronic')

        return True

    def stopCommunication(self):
        """
        stopCommunication disables the cyclic timers for polling necessary relay data.

        :return: True for test purpose
        """
        self.timerTask.stop()
        self.signals.deviceDisconnected.emit('KMTronic')

        return True

    def debugOutput(self, result=None):
        """
        debugOutput writes a nicely formed output for diagnosis in relay environment

        :param result: value from requests get commend
        :return: True for test purpose
        """

        if result is None:
            self.log.info('No valid result')
            return False

        text = result.text.replace('\r\n', ', ')
        reason = result.reason
        status = result.status_code
        url = result.url
        elapsed = result.elapsed

        self.log.trace(f'Result: {url}, {reason}, {status}, {elapsed}, {text}')

        return True

    def getRelay(self, url='/status.xml', debug=True):
        """
        getRelay sets and reads data from the given host ip using the given
        user and password

        :param url: web address of relay box
        :param debug: write extended debug output
        :return: result: return values from web interface of box
        """

        if self.hostaddress is None:
            return None

        if not self.mutexPoll.tryLock():
            return None

        auth = requests.auth.HTTPBasicAuth(self.user, self.password)
        url = f'http://{self.hostaddress}:80{url}'
        result = None

        try:
            result = requests.get(url, auth=auth, timeout=self.TIMEOUT)

        except requests.exceptions.Timeout:
            pass

        except requests.exceptions.ConnectionError:
            pass

        except Exception as e:
            self.log.critical(f'Error in request: {e}')

        if debug:
            self.debugOutput(result=result)

        self.mutexPoll.unlock()
        return result

    def cyclePolling(self):
        """
        cyclePolling reads the status of the relay status of each single relay.
        with success the statusReady single is sent.

        :return: success
        """

        value = self.getRelay('/status.xml', debug=False)

        if value is None:
            return False

        if value.reason != 'OK':
            return False

        lines = value.text.splitlines()

        for line in lines:
            value = re.findall(r'\d', line)
            if not value:
                continue
            value = [int(s) for s in value]
            self.status[value[0] - 1] = value[1]

        self.signals.statusReady.emit()
        self.signals.deviceConnected.emit('KMTronic')

        return True

    def getByte(self, relayNumber=0, state=False):
        """
        getByte generates the right bit mask for setting or resetting the relay mask

        :param relayNumber: relay number
        :param state: state to archive
        :return: bit mask for switching
        """

        byteStat = 0b0
        for i, status in enumerate(self.status):
            if status:
                byteStat = byteStat | 1 << i
        position = 1 << relayNumber
        byteOn = byteStat | position
        byteOff = byteOn & ~position

        if state:
            return byteOn
        else:
            return byteOff

    def pulse(self, relayNumber):
        """
        pulse switches a relay on for one second and off back.

        :param relayNumber: number of relay to be pulsed, counting from 0 onwards
        :return: success
        """

        self.log.debug(f'Pulse relay:{relayNumber}')
        byteOn = self.getByte(relayNumber=relayNumber, state=True)
        byteOff = self.getByte(relayNumber=relayNumber, state=False)
        value1 = self.getRelay(f'/FFE0{byteOn:02X}')
        time.sleep(self.PULSEWIDTH)
        value2 = self.getRelay(f'/FFE0{byteOff:02X}')

        if value1 is None or value2 is None:
            self.log.warning(f'Relay:{relayNumber}')
            return False
        elif value1.reason != 'OK' or value2.reason != 'OK':
            self.log.warning(f'Relay:{relayNumber}')
            return False

        return True

    def switch(self, relayNumber):
        """
        switch toggles the relay status (on, off)

        :param relayNumber: number of relay to be pulsed, counting from 0 onwards
        :return: success
        """

        self.log.debug(f'Switch relay:{relayNumber}')
        value = self.getRelay('/relays.cgi?relay={0:1d}'.format(relayNumber + 1))

        if value is None:
            self.log.warning(f'Relay:{relayNumber}')
            return False
        elif value.reason != 'OK':
            self.log.warning(f'Relay:{relayNumber}')
            return False

        return True

    def set(self, relayNumber, value):
        """
        set toggles the relay status to the desired value (on, off)

        :param relayNumber: number of relay to be pulsed, counting from 0 onwards
        :param value: relay state.
        :return: success
        """

        self.log.debug(f'Set relay:{relayNumber}')
        byteOn = self.getByte(relayNumber=relayNumber, state=value)
        value = self.getRelay(f'/FFE0{byteOn:02X}')

        if value is None:
            self.log.warning(f'Relay:{relayNumber}')
            return False
        elif value.reason != 'OK':
            self.log.warning(f'Relay:{relayNumber}')
            return False

        return True



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
#
# written in python3, (c) 2019-2021 by mworion
#
# Licence APL2.0
#
###########################################################
# standard libraries

# external packages
from PyQt5.QtGui import QVector3D
from PyQt5.Qt3DCore import QEntity
from PyQt5.Qt3DExtras import QCuboidMesh

# local import
from gui.extWindows.simulator.materials import Materials
from gui.extWindows.simulator import tools


class SimulatorTelescope:

    __all__ = ['SimulatorTelescope',
               ]

    def __init__(self, app):
        super().__init__()

        self.app = app
        self.model = {}
        self.modelRoot = None

    def create(self, rEntity, show):
        """

        dict {'name of model': {'parent': }}

        :param rEntity: root of the 3D models
        :param show: root of the 3D models
        :return:
        """

        if self.model:
            self.modelRoot.setParent(None)

        self.model.clear()

        if not show:
            return False

        self.modelRoot = QEntity(rEntity)

        self.model = {
            'mountBase': {
                'parent': None,
                'source': 'mount-base.stl',
                'trans': [0, 0, 1000],
                'mat': Materials().aluminiumS,
            },
            'mountKnobs': {
                'parent': 'mountBase',
                'source': 'mount-base-knobs.stl',
                'mat': Materials().aluminium,
            },
            'lat': {
                'parent': 'mountBase',
                'trans': [0, 0, 70],
                'rot': [0, -90 + 48, 0],
            },
            'mountRa': {
                'parent': 'lat',
                'source': 'mount-ra.stl',
                'trans': [0, 0, -70],
                'mat': Materials().aluminiumS,
            },
            'ra': {
                'parent': 'mountRa',
                'trans': [0, 0, 190],
            },
            'mountDec': {
                'parent': 'ra',
                'source': 'mount-dec.stl',
                'trans': [0, 0, -190],
                'mat': Materials().aluminiumS,
            },
            'mountDecKnobs': {
                'parent': 'ra',
                'source': 'mount-dec-knobs.stl',
                'trans': [0, 0, -190],
                'mat': Materials().aluminium,
            },
            'mountDecWeights': {
                'parent': 'ra',
                'source': 'mount-dec-weights.stl',
                'trans': [0, 0, -190],
                'mat': Materials().stainless,
            },
            'dec': {
                'parent': 'mountDec',
                'trans': [159, 0, 190],
            },
            'mountHead': {
                'parent': 'dec',
                'source': 'mount-head.stl',
                'trans': [-159, 0, -190],
                'mat': Materials().aluminiumS,
            },
            'mountHeadKnobs': {
                'parent': 'dec',
                'source': 'mount-head-knobs.stl',
                'trans': [-159, 0, -190],
                'mat': Materials().aluminium,
            },
            'gem': {
                'parent': 'mountHead',
                'source': [QCuboidMesh(), 100, 60, 10],
                'trans': [159, 0, 338.5],
                'mat': Materials().aluminiumB,
            },
            'gemCorr': {
                'parent': 'gem',
                'scale': [1, 1, 1],
            },
            'otaPlate': {
                'parent': 'gemCorr',
                'source': 'ota-plate.stl',
                'mat': Materials().aluminiumS,
            },
            'otaRing': {
                'parent': 'otaPlate',
                'source': 'ota-ring-s.stl',
                'scale': [1, 1, 1],
                'mat': Materials().aluminiumS,
            },
            'otaTube': {
                'parent': 'otaPlate',
                'source': 'ota-tube-s.stl',
                'scale': [1, 1, 1],
                'mat': Materials().white,
            },
            'otaImagetrain': {
                'parent': 'gemCorr',
                'source': 'ota-imagetrain.stl',
                'scale': [1, 1, 1],
                'mat': Materials().aluminiumS,
            },
            'otaCCD': {
                'parent': 'otaImagetrain',
                'source': 'ota-ccd.stl',
                'mat': Materials().aluminiumB,
            },
            'otaFocus': {
                'parent': 'otaImagetrain',
                'source': 'ota-focus.stl',
                'mat': Materials().aluminiumR,
            },
            'otaFocusTop': {
                'parent': 'otaImagetrain',
                'source': 'ota-focus-top.stl',
                'mat': Materials().white,
            },
        }

        for name in self.model:
            tools.linkModel(self.model, name, self.modelRoot)

        return True

    def updateSettings(self):
        """
        updateSettings resize parts depending on the setting made in the dome tab. likewise
        some transformations have to be reverted as they are propagated through entity linking.

        :return:
        """

        if not self.model:
            return False

        north = self.app.mount.geometry.offNorth * 1000
        east = self.app.mount.geometry.offEast * 1000
        vertical = self.app.mount.geometry.offVert * 1000
        self.model['mountBase']['t'].setTranslation(QVector3D(north, -east, 1000 + vertical))

        if self.app.mount.obsSite.location:
            latitude = self.app.mount.obsSite.location.latitude.degrees
            self.model['lat']['t'].setRotationY(- abs(latitude))

        if self.app.mount.geometry.offGemPlate:
            offPlateOTA = self.app.mount.geometry.offPlateOTA * 1000
            lat = - self.app.mainW.ui.offLAT.value() * 1000

            self.model['gem']['m'].setYExtent(abs(lat) + 80)
            self.model['gem']['t'].setTranslation(QVector3D(159.0, lat / 2, 338.5))
            self.model['gemCorr']['t'].setTranslation(QVector3D(0.0, lat / 2, 0.0))

            scaleRad = (offPlateOTA - 25) / 55
            scaleRad = max(scaleRad, 1)

            self.model['otaRing']['t'].setScale3D(QVector3D(1.0, scaleRad, scaleRad))
            self.model['otaRing']['t'].setTranslation(QVector3D(0.0, 0.0, - 10 * scaleRad + 10))
            self.model['otaTube']['t'].setScale3D(QVector3D(1.0, scaleRad, scaleRad))
            self.model['otaTube']['t'].setTranslation(QVector3D(0.0, 0.0, - 10 * scaleRad + 10))
            self.model['otaImagetrain']['t'].setTranslation(QVector3D(0, 0, 65 * (scaleRad - 1)))

        return True

    def updatePositions(self):
        """
        updateMount moves ra and dec axis according to the values in the mount.

        :return:
        """

        if not self.model:
            return False

        angRA = self.app.mount.obsSite.angularPosRA
        angDEC = self.app.mount.obsSite.angularPosDEC

        if not (angRA and angDEC):
            return False

        self.model['ra']['t'].setRotationX(- angRA.degrees + 90)
        self.model['dec']['t'].setRotationZ(- angDEC.degrees)

        return True



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
#
# written in python3, (c) 2019-2021 by mworion
# Licence APL2.0
#
###########################################################
# standard libraries

# external packages
from PyQt5.QtCore import QUrl
from PyQt5.QtGui import QVector3D, QFont
from PyQt5.Qt3DExtras import QCuboidMesh, QSphereMesh
from PyQt5.Qt3DExtras import QExtrudedTextMesh, QCylinderMesh
from PyQt5.Qt3DRender import QMesh
from PyQt5.Qt3DCore import QEntity, QTransform

# local import


def linkSource(currMod):
    """

    :param currMod:
    :return: mesh
    """

    source = currMod.get('source', None)
    if source:
        if isinstance(source, str):
            mesh = QMesh()
            mesh.setSource(QUrl(f'qrc:/model3D/{source}'))
        elif isinstance(source[0], QCuboidMesh):
            mesh = source[0]
            mesh.setXExtent(source[1])
            mesh.setYExtent(source[2])
            mesh.setZExtent(source[3])
        elif isinstance(source[0], QSphereMesh):
            mesh = source[0]
            mesh.setRadius(source[1])
            mesh.setRings(source[2])
            mesh.setSlices(source[3])
        elif isinstance(source[0], QCylinderMesh):
            mesh = source[0]
            mesh.setLength(source[1])
            mesh.setRadius(source[2])
            mesh.setRings(source[3])
            mesh.setSlices(source[4])
        elif isinstance(source[0], QExtrudedTextMesh):
            mesh = source[0]
            mesh.setDepth(source[1])
            mesh.setFont(QFont())
            mesh.setText(source[3])
        else:
            mesh = None
    else:
        mesh = None

    return mesh


def linkTransform(currMod):
    """

    :param currMod:
    :return: transform
    """

    trans = currMod.get('trans', None)
    rot = currMod.get('rot', None)
    scale = currMod.get('scale', None)

    if trans or rot or scale:
        transform = QTransform()

        if trans and isinstance(trans, list) and len(trans) == 3:
            transform.setTranslation(QVector3D(*trans))

        if rot and isinstance(rot, list) and len(rot) == 3:
            transform.setRotationX(rot[0])
            transform.setRotationY(rot[1])
            transform.setRotationZ(rot[2])

        if scale and isinstance(scale, list) and len(scale) == 3:
            transform.setScale3D(QVector3D(*scale))
    else:
        transform = None

    return transform


def linkMaterial(currMod):
    """

    :param currMod:
    :return: material
    """

    mat = currMod.get('mat', None)
    return mat


def linkModel(model, name, rEntity):
    """

    :param model:
    :param name:
    :param rEntity:
    :return: true for test purpose
    """

    currMod = model[name]

    parent = currMod.get('parent', None)
    if parent and model.get(parent, None):
        currMod['e'] = QEntity(model[parent]['e'])
    else:
        currMod['e'] = QEntity(rEntity)

    mesh = linkSource(currMod)
    if mesh:
        currMod['e'].addComponent(mesh)
        currMod['m'] = mesh

    transform = linkTransform(currMod)
    if transform:
        currMod['e'].addComponent(transform)
        currMod['t'] = transform

    material = linkMaterial(currMod)
    if material:
        currMod['mat'] = material
        currMod['e'].addComponent(material)

    return True



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
#
# written in python3, (c) 2019-2021 by mworion
#
# Licence APL2.0
#
###########################################################
# standard libraries
import logging
import os
import json
import random
import csv

# external packages
import numpy as np
import skyfield.api
from scipy.spatial import distance

# local imports
from base import transform

__all__ = ['HaDecToAltAz',
           'DataPoint',
           ]


def HaDecToAltAz(ha, dec, lat):
    """
    HaDecToAltAz is derived from http://www.stargazing.net/kepler/altaz.html

    :param ha: hour angle in [h]
    :param dec: declination in [deg]
    :param lat: latitude of observer
    :return: altitude and azimuth
    """

    ha = (ha * 360 / 24 + 360.0) % 360.0
    dec = np.radians(dec)
    ha = np.radians(ha)
    lat = np.radians(lat)
    alt = np.arcsin(np.sin(dec) * np.sin(lat) + np.cos(dec) * np.cos(lat) * np.cos(ha))
    value = (np.sin(dec) - np.sin(alt) * np.sin(lat)) / (np.cos(alt) * np.cos(lat))
    value = np.clip(value, -1.0, 1.0)
    A = np.arccos(value)
    if np.sin(ha) >= 0.0:
        az = 2 * np.pi - A
    else:
        az = A
    az = np.degrees(az)
    alt = np.degrees(alt)
    return alt, az


class DataPoint(object):
    """
    The class Data inherits all information and handling of modeldata data and other
    attributes. this includes horizon data, model points data and their persistence

        >>> data = DataPoint(
        >>>                  app=None,
        >>>                  )
    """

    __all__ = ['DataPoint']

    log = logging.getLogger(__name__)

    # data for generating greater circles, dec and step only for east, west is reversed
    DEC = {'min': [-15, 0, 15, 30, 45, 60, 75],
           'norm': [-15, 0, 15, 30, 45, 60, 75],
           'med': [-15, -5, 5, 15, 25, 40, 55, 70, 85],
           'max': [-15, -5, 5, 15, 25, 35, 45, 55, 65, 75, 85],
           }
    STEP = {'min': [15, -15, 15, -15, 15, -30, 30],
            'norm': [10, -10, 10, -10, 10, -20, 20],
            'med': [10, -10, 10, -10, 10, -10, 10, -30, 30],
            'max': [10, -10, 10, -10, 10, -10, 10, -10, 10, -30, 30],
            }
    START = {'min': [-120, -5, -120, -5, -120, -5, -120,
                     5, 120, 5, 120, 5, 120, 5],
             'norm': [-120, -5, -120, -5, -120, -5, -120,
                      5, 120, 5, 120, 5, 120, 5],
             'med': [-120, -5, -120, -5, -120, -5, -120, -5, -120,
                     5, 120, 5, 120, 5, 120, 5, 120, 5],
             'max': [-120, -5, -120, -5, -120, -5, -120, -5, -120, -5, -120,
                     5, 120, 5, 120, 5, 120, 5, 120, 5, 120, 5],
             }
    STOP = {'min': [0, -120, 0, -120, 0, -120, 0,
                    120, 0, 120, 0, 120, 0, 120],
            'norm': [0, -120, 0, -120, 0, -120, 0,
                     120, 0, 120, 0, 120, 0, 120],
            'med': [0, -120, 0, -120, 0, -120, 0, -120, 0,
                    120, 0, 120, 0, 120, 0, 120, 0, 120, 0],
            'max': [0, -120, 0, -120, 0, -120, 0, -120, 0, -120, 0,
                    120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0],
            }

    def __init__(self,
                 app=None,
                 ):

        self.app = app
        self.configDir = app.mwGlob['configDir']
        self._horizonP = list()
        self._buildP = list()

    @property
    def horizonP(self):
        return self._horizonP

    @horizonP.setter
    def horizonP(self, value):
        if not isinstance(value, list):
            self._horizonP.clear()
            return

        if not all([isinstance(x, tuple) for x in value]):
            self.log.info('Malformed value: {0}'.format(value))
            self._horizonP.clear()
            return

        self._horizonP.clear()
        self._horizonP += value
        return

    @property
    def buildP(self):
        return self._buildP

    @buildP.setter
    def buildP(self, value):
        if not isinstance(value, list):
            self._buildP = list()
            return

        if not all([isinstance(x, tuple) for x in value]):
            self.log.info('Malformed value: {0}'.format(value))
            self._buildP = list()
            return

        self._buildP = value

    def addBuildP(self, value=None, position=None):
        """
        addBuildP extends the list of modeldata points. the new point could be added at the end
        of the list (default) or in any location in the list.

        :param value: value to be inserted
        :param position: position in list
        :return:
        """

        if value is None:
            return False

        if not isinstance(value, tuple):
            self.log.info('malformed value: {0}'.format(value))
            return False

        if len(value) != 3:
            self.log.info('malformed value: {0}'.format(value))
            return False

        if position is None:
            position = len(self._buildP)

        if not isinstance(position, (int, float)):
            self.log.info('malformed position: {0}'.format(position))
            return False

        if self.app.mount.setting.horizonLimitHigh is not None:
            high = self.app.mount.setting.horizonLimitHigh

        else:
            high = 90

        if self.app.mount.setting.horizonLimitLow is not None:
            low = self.app.mount.setting.horizonLimitLow

        else:
            low = 0

        if value[0] > high:
            return False

        if value[0] < low:
            return False

        position = int(position)
        position = min(len(self._buildP), position)
        position = max(0, position)
        self._buildP.insert(position, value)

        return True

    def delBuildP(self, position):
        """
        delBuildP deletes one point from the modeldata points list at the given index.

        :param position:
        :return:
        """

        if not isinstance(position, (int, float)):
            self.log.info('malformed position: {0}'.format(position))
            return False

        position = int(position)

        if position < 0 or position > len(self._buildP) - 1:
            self.log.info('invalid position: {0}'.format(position))
            return False

        self._buildP.pop(position)

        return True

    def clearBuildP(self):
        self._buildP.clear()

    def setStatusBuildP(self, number, status):
        """

        :param number:
        :param status:
        :return: True for test purpose
        """
        if number < 0:
            return False

        if number > len(self._buildP) - 1:
            return False

        alt, az, _ = self._buildP[number]
        self._buildP[number] = (alt, az, status)

        return True

    def addHorizonP(self, value=None, position=None):
        """
        addHorizonP extends the list of modeldata points. the new point could be added at the
        end of the list (default) or in any location in the list.

        :param value:
        :param position:
        :return:
        """

        if value is None:
            return False

        if not isinstance(value, tuple):
            self.log.info('malformed value: {0}'.format(value))
            return False

        if len(value) != 2:
            self.log.info('malformed value: {0}'.format(value))
            return False

        if position is None:
            position = len(self.horizonP)

        if not isinstance(position, (int, float)):
            self.log.info('malformed position: {0}'.format(position))
            return False

        position = int(position)
        position = min(len(self._horizonP), position)
        position = max(0, position)
        self._horizonP.insert(position, value)

        return True

    def delHorizonP(self, position):
        """
        delHorizonP deletes one point from the modeldata points list at the given index.

        :param position:
        :return:
        """

        if not isinstance(position, (int, float)):
            self.log.info('malformed position: {0}'.format(position))
            return False

        position = int(position)

        if position < 0 or position > len(self._horizonP):
            self.log.info('invalid position: {0}'.format(position))
            return False

        self._horizonP.pop(position)

        return True

    def clearHorizonP(self):
        self._horizonP.clear()

    @staticmethod
    def isCloseHorizonLine(point, margin, horizonI):
        """
        see answers
        https://codereview.stackexchange.com/questions
        /28207/finding-the-closest-point-to-a-list-of-points
        :param point:
        :param margin:
        :param horizonI:
        :return:
        """

        pointRef = np.asarray([point[1], point[0]])
        closest_index = distance.cdist([pointRef], horizonI).argmin()
        pointClose = horizonI[closest_index]
        val = np.sqrt(np.sum((pointRef - pointClose)**2))

        if val < margin:
            return True

        else:
            return False

    def isAboveHorizon(self, point):
        """
        isAboveHorizon calculates for a given point the relationship to the actual horizon
        and determines if this point is above the horizon line. for that there will be a
        linear interpolation for the horizon line points.

        :param point:
        :return:
        """
        if point[1] > 360:
            point = (point[0], 360)

        if point[1] < 0:
            point = (point[0], 0)

        x = range(0, 361)

        if self.horizonP:
            xRef = [i[1] for i in self.horizonP]
            yRef = [i[0] for i in self.horizonP]

        else:
            xRef = [0]
            yRef = [0]

        y = np.interp(x, xRef, yRef)

        if point[0] > y[int(point[1])]:
            return True

        else:
            return False

    def isCloseMeridian(self, point):
        """
        isCloseMeridian check if a point is close to meridian slew / track limits limit

        :param point:
        :return: status
        """
        slew = self.app.mount.setting.meridianLimitSlew
        track = self.app.mount.setting.meridianLimitTrack

        if slew is None or track is None:
            return False

        value = max(slew, track)

        lower = 180 - value
        upper = 180 + value

        if lower < point[1] < upper:
            return True

        else:
            return False

    def deleteBelowHorizon(self):
        """
        deleteBelowHorizon deletes all points which are below horizon line

        :return: true for test purpose
        """
        self._buildP = [x for x in self._buildP if self.isAboveHorizon(x)]
        return True

    def deleteCloseMeridian(self):
        """
        deleteCloseMeridian deletes all points which are too close to the meridian and
        therefore need to flip all the time

        :return: true for test purpose
        """
        self._buildP = [x for x in self._buildP if not self.isCloseMeridian(x)]
        return True

    def deleteCloseHorizonLine(self, m):
        """
        :param m: margin to horizon line in degrees
        :return:
        """
        if not self.horizonP:
            return

        azH = [x[1] for x in self.horizonP]
        altH = [x[0] for x in self.horizonP]
        azI = range(0, 361, 1)
        altI = np.interp(azI, azH, altH)
        horizonI = np.asarray([[x, y] for x, y in zip(azI, altI)])
        self._buildP = [x for x in self._buildP if not self.isCloseHorizonLine(x, m, horizonI)]

        return True

    def sort(self, eastwest=False, highlow=False, pierside=None):
        """
        :param eastwest: flag if to be sorted east - west
        :param highlow:  flag if sorted high low altitude
        :param pierside:  start pierside sorting with this position
        :return: true for test purpose
        """

        if eastwest and highlow:
            return False

        if not eastwest and not highlow and pierside is None:
            return False

        east = [x for x in self._buildP if x[1] <= 180]
        west = [x for x in self._buildP if x[1] > 180]

        if eastwest:
            east = sorted(east, key=lambda x: -x[1])
            west = sorted(west, key=lambda x: -x[1])

        if highlow:
            east = sorted(east, key=lambda x: -x[0])
            west = sorted(west, key=lambda x: -x[0])

        if pierside == 'E':
            self._buildP = west + east

        else:
            self._buildP = east + west

        return True

    def loadJSON(self, fileName, ext):
        """

        :param fileName: name of file to be handled
        :param ext: extension of file to be handled
        :return: value: loaded data
        """
        fileName = self.configDir + '/' + fileName + ext

        if not os.path.isfile(fileName):
            return None

        try:
            with open(fileName, 'r') as handle:
                value = json.load(handle)

        except Exception as e:
            self.log.info('Cannot BPTS load: {0}, error: {1}'.format(fileName, e))
            value = None

        else:
            value = [tuple(x) for x in value]

        return value

    def loadCSV(self, fileName, ext, delimiter=','):
        """

        :param fileName: name of file to be handled
        :param ext: extension of file to be handled
        :param delimiter: delimiter of file to be handled
        :return: value: loaded data
        """
        fileName = self.configDir + '/' + fileName + ext
        if not os.path.isfile(fileName):
            return None

        try:
            value = []
            with open(fileName, 'r') as handle:
                reader = csv.reader(handle, delimiter=delimiter)
                for x in reader:
                    x = x[:2]
                    convertedX = [float(val) for val in x]
                    if delimiter == ',':
                        value.append(convertedX)
                    else:
                        value.append(reversed(convertedX))
        except Exception as e:
            self.log.info('Cannot CSV load: {0}, error: {1}'.format(fileName, e))
            value = None
        else:
            value = [tuple(x) for x in value]
        return value

    @staticmethod
    def checkFormat(value):
        if not isinstance(value, list):
            return False

        if not all([isinstance(x, tuple) for x in value]):
            return False

        if not all([len(x) == 2 for x in value]):
            return False

        return True

    def loadBuildP(self, fileName=None, ext='.bpts', keep=False):
        """
        loadBuildP loads a modeldata pints file and stores the data in the buildP list.
        necessary conversion are made.

        :param fileName: name of file to be handled
        :param ext: load extension type
        :param keep:
        :return: success
        """

        if fileName is None:
            return False

        value = None

        if ext == '.csv':
            value = self.loadCSV(fileName, ext)

        elif ext == '.bpts':
            value = self.loadJSON(fileName, ext)

        elif ext == '.txt':
            value = self.loadCSV(fileName, ext, delimiter=':')

        if value is None:
            return False

        suc = self.checkFormat(value)

        if not suc:
            self.clearBuildP()
            return False

        points = [(x[0], x[1], True) for x in value]

        if keep:
            self._buildP += points

        else:
            self._buildP = points

        # backup solution
        if csv:
            self.saveBuildP(fileName=fileName)

        return True

    def saveBuildP(self, fileName=None):
        """
        saveBuildP saves the actual modeldata points list in a file in json dump format

        :param fileName: name of file to be handled
        :return: success
        """

        if fileName is None:
            return False

        fileName = self.configDir + '/' + fileName + '.bpts'

        points = [(x[0], x[1]) for x in self.buildP]

        with open(fileName, 'w') as handle:
            json.dump(points,
                      handle,
                      indent=4)
        return True

    def loadHorizonP(self, fileName=None, ext='.hpts'):
        """
        loadHorizonP loads a modeldata pints file and stores the data in the buildP list.
        necessary conversion are made.

        :param fileName: name of file to be handled
        :param ext: load extension type
        :return: success
        """

        if fileName is None:
            return False

        value = None
        if ext == '.csv':
            value = self.loadCSV(fileName, ext)

        elif ext == '.hpts':
            value = self.loadJSON(fileName, ext)

        elif ext == '.txt':
            value = self.loadCSV(fileName, ext, delimiter=':')

        suc = self.checkFormat(value)

        if not suc:
            self.clearHorizonP()
            return False

        self.horizonP = value

        # backup solution
        if csv:
            self.saveHorizonP(fileName=fileName)

        return True

    def saveHorizonP(self, fileName=None):
        """
        saveHorizonP saves the actual modeldata points list in a file in json dump format

        :param fileName: name of file to be handled
        :return: success
        """

        if fileName is None:
            return False

        fileName = self.configDir + '/' + fileName + '.hpts'

        with open(fileName, 'w') as handle:
            json.dump(self.horizonP,
                      handle,
                      indent=4)

        return True

    def genHaDecParams(self, selection):
        """
        genHaDecParams selects the parameters for generating the boundaries for next
        step processing greater circles. the parameters are sorted for different targets
        actually for minimum slew distance between the points. defined is only the east
        side of data, the west side will be mirrored to the east one.

        :param selection: type of model we would like to use
        :return: yield tuple of dec value and step, start and stop for range
        """

        if selection not in self.DEC or selection not in self.STEP:
            return

        eastDec = self.DEC[selection]
        westDec = list(reversed(eastDec))
        decL = eastDec + westDec

        eastStepL = self.STEP[selection]
        westStepL = list(reversed(eastStepL))
        stepL = eastStepL + westStepL
        startL = self.START[selection]
        stopL = self.STOP[selection]

        for dec, step, start, stop in zip(decL, stepL, startL, stopL):
            yield dec, step, start, stop

    def genGreaterCircle(self, selection='norm', keep=False):
        """
        genGreaterCircle takes the generated boundaries for the rang routine and
        transforms ha, dec to alt az. reasonable values for the alt az values
        are 5 to 85 degrees.

        :param selection:
        :param keep:
        :return: yields alt, az tuples which are above horizon
        """

        if not self.app.mount.obsSite.location:
            return False

        if not keep:
            self.clearBuildP()

        lat = self.app.mount.obsSite.location.latitude.degrees
        for dec, step, start, stop in self.genHaDecParams(selection):
            for ha in range(start, stop, step):
                alt, az = HaDecToAltAz(ha / 10, dec, lat)
                # only values with above horizon = 0

                if 5 <= alt <= 85 and 2 < az < 358:
                    alt += random.uniform(-2, 2)
                    self.addBuildP((alt, az, True))

        return True

    @staticmethod
    def genGridGenerator(eastAlt, westAlt, minAz, stepAz, maxAz):
        """
        genGridGenerator generates the point values out of the given ranges of altitude
        and azimuth

        :param eastAlt:
        :param westAlt:
        :param minAz:
        :param stepAz:
        :param maxAz:
        :return:
        """

        for i, alt in enumerate(eastAlt):
            if i % 2:
                for az in range(minAz, 180, stepAz):
                    yield(alt, az, True)
            else:
                for az in range(180 - minAz, 0, -stepAz):
                    yield(alt, az, True)

        for i, alt in enumerate(westAlt):
            if i % 2:
                for az in range(180 + minAz, 360, stepAz):
                    yield(alt, az, True)
            else:
                for az in range(maxAz, 180, -stepAz):
                    yield(alt, az, True)

    def genGrid(self, minAlt=5, maxAlt=85, numbRows=5, numbCols=6, keep=False):
        """
        genGrid generates a grid of points and transforms ha, dec to alt az. with given
        limits in alt, the min and max will be used as a hard condition. on az there is
        not given limit, therefore a split over the whole space (omitting the meridian)
        is done. the simplest way to avoid hitting the meridian is to enforce the number
        of cols to be a factor of 2. reasonable values for the grid are 5 to 85 degrees.
        defined is only the east side of data, the west side will be mirrored to the
        east one.
            the number of rows is 2 < x < 8
            the number of columns is 2 < x < 15

        :param minAlt: altitude min
        :param maxAlt: altitude max
        :param numbRows: numbRows
        :param numbCols: numbCols
        :param keep:
        :return: yields alt, az tuples which are above horizon
        """

        if not 5 <= minAlt <= 85:
            return False

        if not 5 <= maxAlt <= 85:
            return False

        if not maxAlt > minAlt:
            return False

        if not 1 < numbRows < 9:
            return False

        if not 1 < numbCols < 16:
            return False

        if numbCols % 2:
            return False

        minAlt = int(minAlt)
        maxAlt = int(maxAlt)
        numbCols = int(numbCols)
        numbRows = int(numbRows)

        stepAlt = int((maxAlt - minAlt) / (numbRows - 1))
        eastAlt = list(range(minAlt, maxAlt + 1, stepAlt))
        westAlt = list(reversed(eastAlt))

        stepAz = int(360 / numbCols)
        minAz = int(180 / numbCols)
        maxAz = 360 - minAz

        if not keep:
            self.clearBuildP()

        for point in self.genGridGenerator(eastAlt, westAlt, minAz, stepAz, maxAz):
            self.addBuildP(point)

        return True

    def genAlign(self, altBase=30, azBase=10, numberBase=3, keep=False):
        """
        genAlign generates a number of initial points for the first step of modeling. it
        adjusts the first align point in a matter, that the starting point is the closest
        to az = 0.

        :param altBase:
        :param azBase:
        :param numberBase:
        :param keep:
        :return: yields alt, az tuples which are above horizon
        """

        if not 5 <= altBase <= 85:
            return False

        if not 2 < numberBase < 11:
            return False

        if not 0 <= azBase < 360:
            return False

        stepAz = int(360 / numberBase)
        altBase = int(altBase)
        azBase = int(azBase) % stepAz
        numberBase = int(numberBase)

        if not keep:
            self.clearBuildP()

        for i in range(0, numberBase):
            az = azBase + i * stepAz
            self.addBuildP((altBase, az % 360, True))

        return True

    def generateCelestialEquator(self):
        """
        generateCelestialEquator calculates a line for greater circles like a celestial
        equator for showing the paths in the hemisphere window.

        :return: celestial equator
        """

        celestialEquator = list()

        if not self.app.mount.obsSite.location:
            return celestialEquator

        lat = self.app.mount.obsSite.location.latitude.degrees

        for dec in range(-15, 90, 15):
            for ha in range(- 119, 120, 2):
                az, alt = HaDecToAltAz(ha / 10, dec, lat)
                if alt > 0:
                    celestialEquator.append((az, alt))

        for ha in range(-115, 120, 10):
            for dec in range(- 90, 90, 2):
                az, alt = HaDecToAltAz(ha / 10, dec, lat)
                if alt > 0:
                    celestialEquator.append((az, alt))

        return celestialEquator

    def generateDSOPath(self, ra=0, dec=0, timeJD=0, location=None,
                        numberPoints=0, duration=0, timeShift=0, keep=False):
        """
        generateDSOPath calculates a list of model points along the desired path beginning
        at ra, dec coordinates, which is in time duration hours long and consists of
        numberPoints model points. TimeShift moves the pearl of points to an earlier or
        later point in time.

        :param ra:
        :param dec:
        :param timeJD:
        :param location:
        :param numberPoints:
        :param duration:
        :param timeShift:
        :param keep:
        :return: True for test purpose
        """

        if numberPoints < 1:
            return False

        if duration == 0:
            return False

        if location is None:
            return False

        numberPoints = int(numberPoints)

        if not keep:
            self.clearBuildP()

        for i in range(0, numberPoints):
            startPoint = ra.hours - i * duration / numberPoints - timeShift
            raCalc = skyfield.api.Angle(hours=startPoint)
            az, alt = transform.J2000ToAltAz(raCalc, dec, timeJD, location)
            if alt.degrees > 0:
                self.addBuildP((alt.degrees, az.degrees % 360, True))

        return True

    def generateGoldenSpiral(self, numberPoints, keep=False):
        """
        based on the evaluations and implementation of CR Drost from 17-05-24 found at:
        https://stackoverflow.com/questions/9600801/evenly-distributing-n-points-on-a-sphere
        the implementation of an equally distributed points cloud over on half of the
        hemisphere.

        :param numberPoints:
        :param keep:
        :return: true for test purpose
        """

        if numberPoints < 50:
            return False

        if not keep:
            self.clearBuildP()

        indices = np.arange(0, numberPoints, dtype=float) + 0.5
        phi = np.arccos(1 - 2 * indices / numberPoints)
        theta = np.pi * (1 + 5 ** 0.5) * indices

        altitude = 90 - np.degrees(phi)
        azimuth = np.degrees(theta) % 360

        for alt, az in zip(altitude, azimuth):
            if alt > 0:
                self.addBuildP((alt, az, True))

        return True



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
#
# written in python3, (c) 2019-2021 by mworion
# Licence APL2.0
#
###########################################################
# standard libraries
import logging
import os

# external packages
import PyQt5
from astropy.io import fits
import numpy as np

# local imports
from mountcontrol.convert import convertToAngle
from base import tpool
from logic.astrometry.astrometryNET import AstrometryNET
from logic.astrometry.astrometryASTAP import AstrometryASTAP


class AstrometrySignals(PyQt5.QtCore.QObject):
    """
    """

    __all__ = ['AstrometrySignals']

    done = PyQt5.QtCore.pyqtSignal(object)
    result = PyQt5.QtCore.pyqtSignal(object)
    message = PyQt5.QtCore.pyqtSignal(object)

    serverConnected = PyQt5.QtCore.pyqtSignal()
    serverDisconnected = PyQt5.QtCore.pyqtSignal(object)
    deviceConnected = PyQt5.QtCore.pyqtSignal(object)
    deviceDisconnected = PyQt5.QtCore.pyqtSignal(object)


class Astrometry:
    """
    the class Astrometry inherits all information and handling of astrometry.net
    handling

    Keyword definitions could be found under
        https://fits.gsfc.nasa.gov/fits_dictionary.html

        >>> astrometry = Astrometry(app=None)
    """

    __all__ = ['Astrometry',
               ]

    log = logging.getLogger(__name__)

    def __init__(self, app):
        self.app = app
        self.tempDir = app.mwGlob['tempDir']
        self.threadPool = app.threadPool
        self.signals = AstrometrySignals()

        self.data = {}
        self.defaultConfig = {'framework': '',
                              'frameworks': {}}
        self.framework = ''
        self.run = {
            'astrometry': AstrometryNET(self),
            'astap': AstrometryASTAP(self),
        }
        for fw in self.run:
            self.defaultConfig['frameworks'].update(self.run[fw].defaultConfig)

        self.mutexSolve = PyQt5.QtCore.QMutex()

    def readFitsData(self, fitsPath):
        """
        readFitsData reads the fits file with the image and tries to get some
        key fields out of the header for preparing the solver. if there is the
        need for understanding more FITS header data, it should be integrated
        in this method.

        :param fitsPath: fits file with image data
        :return: raHint, decHint, scaleHint
        """
        with fits.open(fitsPath) as fitsHDU:
            fitsHeader = fitsHDU[0].header
            scaleHint = float(fitsHeader.get('SCALE', 0))
            ra = fitsHeader.get('RA', 0)
            dec = fitsHeader.get('DEC', 0)
            raHint = convertToAngle(ra, isHours=True)
            decHint = convertToAngle(dec, isHours=False)

        self.log.debug(f'Header RA: {raHint} ({ra}), DEC: {decHint} ({dec})'
                       f', Scale: {scaleHint}')

        return raHint, decHint, scaleHint, ra, dec

    @staticmethod
    def calcAngleScaleFromWCS(wcsHeader=None):
        """
        calcAngleScaleFromWCS as the name says. important is to use the numpy
        arctan2 function, because it handles the zero points and extend the
        calculation back to the full range from -pi to pi

        :return: angle in degrees and scale in arc second per pixel (app) and
                 status if image is mirrored (not rotated for 180 degrees because
                 of the mount flip)
        """
        CD11 = wcsHeader.get('CD1_1', 0)
        CD12 = wcsHeader.get('CD1_2', 0)
        CD21 = wcsHeader.get('CD2_1', 0)
        CD22 = wcsHeader.get('CD2_2', 0)

        mirrored = (CD11 * CD22 - CD12 * CD21) < 0

        angleRad = np.arctan2(CD12, CD11)
        angle = np.degrees(angleRad)
        scale = CD11 / np.cos(angleRad) * 3600

        return angle, scale, mirrored

    def getSolutionFromWCS(self, fitsHeader=None, wcsHeader=None, updateFits=False):
        """
        getSolutionFromWCS reads the wcs fits file and uses the data in the
        header containing the wcs data and returns the basic data needed.
        in addition it embeds it to the given fits file with image. it removes
        all entries starting with some keywords given in selection. we starting
        with HISTORY

        CRVAL1 and CRVAL2 give the center coordinate as right ascension and
        declination or longitude and latitude in decimal degrees.

        the difference is calculated as real coordinate (= plate solved
        coordinate) and mount reported coordinate (= including the errors) and
        set positive in this case.

        we have to take into account if the mount is on the other pierside, the
        image taken will be upside down and the angle will reference a 180
        degrees turned image. this will lead to the negative error value (sign
        will change)

        :param fitsHeader:
        :param wcsHeader:
        :param updateFits:
        :return: ra in hours, dec in degrees, angle in degrees,
                 scale in arcsec/pixel
                 error in arcsec and flag if image is flipped
        """

        self.log.debug(f'wcs header: [{wcsHeader}]')
        raJ2000 = convertToAngle(wcsHeader.get('CRVAL1'), isHours=True)
        decJ2000 = convertToAngle(wcsHeader.get('CRVAL2'), isHours=False)

        angle, scale, mirrored = self.calcAngleScaleFromWCS(wcsHeader=wcsHeader)

        raMount = convertToAngle(fitsHeader.get('RA'), isHours=True)
        decMount = convertToAngle(fitsHeader.get('DEC'), isHours=False)

        deltaRA = (raJ2000._degrees - raMount._degrees) * 3600
        deltaDEC = (decJ2000.degrees - decMount.degrees) * 3600
        error = np.sqrt(np.square(deltaRA) + np.square(deltaDEC))

        solve = {
            'raJ2000S': raJ2000,
            'decJ2000S': decJ2000,
            'errorRA_S': deltaRA,
            'errorDEC_S': deltaDEC,
            'angleS': angle,
            'scaleS': scale,
            'errorRMS_S': error,
            'mirroredS': mirrored,
        }

        if not updateFits:
            return solve, fitsHeader

        fitsHeader.append(('SCALE', solve['scaleS'], 'MountWizzard4'))
        fitsHeader.append(('PIXSCALE', solve['scaleS'], 'MountWizzard4'))
        fitsHeader.append(('ANGLE', solve['angleS'], 'MountWizzard4'))
        fitsHeader.append(('MIRRORED', solve['mirroredS'], 'MountWizzard4'))

        fitsHeader.extend(wcsHeader,
                          unique=True,
                          update=True)

        # remove polynomial coefficients keys if '-SIP' is not selected in
        # CTYPE1 and CTYPE2 this might occur, if you solve a fits file a second
        # time with another solver

        if 'CTYPE1' not in fitsHeader or 'CTYPE2' not in fitsHeader:
            return solve, fitsHeader
        if '-SIP' in fitsHeader['CTYPE1'] and '-SIP' in fitsHeader['CTYPE2']:
            return solve, fitsHeader

        for key in list(fitsHeader.keys()):
            if key.startswith('A_'):
                del fitsHeader[key]
            elif key.startswith('B_'):
                del fitsHeader[key]
            elif key.startswith('AP_'):
                del fitsHeader[key]
            elif key.startswith('BP_'):
                del fitsHeader[key]

        return solve, fitsHeader

    def solveClear(self):
        """
        the cyclic or long lasting tasks for solving the image should not run
        twice for the same data at the same time. so there is a mutex to prevent
        his behaviour.

        :return: true for test purpose
        """
        if self.framework not in self.run:
            return False

        solver = self.run[self.framework]
        self.mutexSolve.unlock()
        self.signals.done.emit(solver.result)
        self.signals.message.emit('')
        return True

    def solveThreading(self, fitsPath='', raHint=None, decHint=None, scaleHint=None,
                       updateFits=False):
        """
        solveThreading is the wrapper for doing the solve process in a
        threadpool environment of Qt. Otherwise the HMI would be stuck all the
        time during solving. it is done with an securing mutex to avoid starting
        solving twice. to solveClear is the partner of solve Threading

        :param fitsPath: full path to the fits image file to be solved
        :param raHint:  ra dest to look for solve in J2000
        :param decHint:  dec dest to look for solve in J2000
        :param scaleHint:  scale to look for solve in J2000
        :param updateFits: flag, if the results should be written to the
                           original file
        :return: success
        """

        if self.framework not in self.run:
            return False

        solver = self.run[self.framework]

        if not os.path.isfile(fitsPath):
            self.signals.done.emit(solver.result)
            return False

        if not self.mutexSolve.tryLock():
            self.log.debug('overrun in solve threading')
            self.signals.done.emit(solver.result)
            return False

        self.signals.message.emit('solving')
        worker = tpool.Worker(solver.solve,
                              fitsPath=fitsPath,
                              raHint=raHint,
                              decHint=decHint,
                              scaleHint=scaleHint,
                              updateFits=updateFits,
                              )
        worker.signals.finished.connect(self.solveClear)
        self.threadPool.start(worker)
        return True

    def abort(self):
        """
        :return:
        """
        if self.framework not in self.run:
            return False

        solver = self.run[self.framework]
        suc = solver.abort()
        return suc

    def checkAvailability(self):
        """
        :return: list of available solutions
        """
        if self.framework not in self.run:
            return False, False

        val = self.run[self.framework].checkAvailability()
        return val

    def startCommunication(self, loadConfig=False):
        """
        :param loadConfig:
        :return: True for test purpose
        """
        self.signals.serverConnected.emit()
        sucApp, sucIndex = self.checkAvailability()
        name = self.run[self.framework].deviceName
        if sucApp and sucIndex:
            self.signals.deviceConnected.emit(name)
            self.app.message.emit(f'ASTROMETRY found:    [{name}]', 0)

        return True

    def stopCommunication(self):
        """
        :return: true for test purpose
        """
        name = self.run[self.framework].deviceName
        self.signals.serverDisconnected.emit({name: 0})
        self.signals.deviceDisconnected.emit(name)
        self.app.message.emit(f'ASTROMETRY remove:   [{name}]', 0)
        return True
